/** \file
\brief Obsuga plikw konfiguracyjnych
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Config \n
Elementy moduu: \ref code_config
*/
#include "Base.hpp"
#include <typeinfo>
#include <map>
#include "Error.hpp"
#include "Stream.hpp"
#include "Files.hpp"
#include "Tokenizer.hpp"
#include "Config.hpp"


namespace common
{

//HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH
// Klasa Config_pimpl

typedef shared_ptr<Item> ITEM_SHARED_PTR;
typedef std::map<tstring, ITEM_SHARED_PTR> ITEM_MAP;

class Config_pimpl
{
public:
	ITEM_MAP ItemMap;

	void ReadConfig(Tokenizer &T, bool NestedConfig);
	void WriteConfig(TCharWriter &W, uint Level);
};

void Config_pimpl::ReadConfig(Tokenizer &T, bool NestedConfig)
{
	tstring Key;
	for (;;)
	{
		// To jest config wewntrzny - zakoczeniem bdzie '}'
		if (NestedConfig)
		{
			if (T.GetToken() == Tokenizer::TOKEN_SYMBOL && T.GetChar() == _T('}'))
			{
				T.Next();
				break;
			}
		}
		// To nie jest config wewntrzny - zakoczeniem bdzie koniec pliku
		else
		{
			if (T.GetToken() == Tokenizer::TOKEN_EOF)
				break;
		}

		// Klucz
		T.AssertToken(Tokenizer::TOKEN_IDENTIFIER);
		Key = T.GetString();
		T.Next();

		// Jest znak rwnoci
		if (T.GetToken() == Tokenizer::TOKEN_SYMBOL && T.GetChar() == _T('='))
		{
			T.Next();

			// '{' - wektor
			if (T.GetToken() == Tokenizer::TOKEN_SYMBOL && T.GetChar() == _T('{'))
			{
				T.Next();

				tstring V;
				shared_ptr<List> SubList(new List);
				if (ItemMap.insert(std::make_pair(Key, SubList)).second == false)
					throw Error(_T("Cannot add list with name: ") + Key);

				while (!T.QuerySymbol(_T('}')))
				{
					if (T.GetToken() == Tokenizer::TOKEN_IDENTIFIER ||
						T.GetToken() == Tokenizer::TOKEN_INTEGER || T.GetToken() == Tokenizer::TOKEN_FLOAT ||
						T.GetToken() == Tokenizer::TOKEN_CHAR || T.GetToken() == Tokenizer::TOKEN_STRING)
					{
						SubList->Data.push_back(T.GetString());
						T.Next();
					}
					else
						T.CreateError();

					// Jeli to nie koniec listy, konieczny przecinek.
					// Jeli koniec, przecinek opcjonalny.
					if (!T.QuerySymbol(_T('}')) && !T.QuerySymbol(_T(',')))
						T.CreateError(_T("Expected ',' or '}'."));
					if (T.QuerySymbol(_T(',')))
						T.Next();
				}
				T.Next();
			}
			// Warto
			else if (
				T.GetToken() == Tokenizer::TOKEN_IDENTIFIER ||
				T.GetToken() == Tokenizer::TOKEN_INTEGER || T.GetToken() == Tokenizer::TOKEN_FLOAT ||
				T.GetToken() == Tokenizer::TOKEN_CHAR || T.GetToken() == Tokenizer::TOKEN_STRING)
			{
				shared_ptr<Value> SubValue(new Value(T.GetString()));
				if (ItemMap.insert(std::make_pair(Key, SubValue)).second == false)
					throw Error(_T("Cannot add value with name: ") + Key);

				T.Next();
			}
		}
		// Podconfig
		else if (T.GetToken() == Tokenizer::TOKEN_SYMBOL && T.GetChar() == _T('{'))
		{
			T.Next();

			shared_ptr<Config> SubConfig(new Config);
			if (ItemMap.insert(std::make_pair(Key, SubConfig)).second == false)
				throw Error(_T("Cannot add subconfig with name: ") + Key);
			SubConfig->pimpl->ReadConfig(T, true);
		}
		else
			T.CreateError();
	}
}

void Config_pimpl::WriteConfig(TCharWriter &W, uint Level)
{
	for (ITEM_MAP::iterator it = ItemMap.begin(); it != ItemMap.end(); ++it)
	{
		// Wcicie
		for (uint i = 0; i < Level; i++)
			W.WriteChar(_T('\t'));
		// Nazwa
		W.WriteString(it->first);

		// Jeli to podkonfiguracja
		if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
		{
			W.WriteChar(_T(' '));
			W.WriteChar(_T('{'));
			W.WriteString(EOL);
			SubConfig->pimpl->WriteConfig(W, Level + 1);
			// Wcicie
			for (uint i = 0; i < Level; i++)
				W.WriteChar(_T('\t'));
			W.WriteChar(_T('}'));
			W.WriteString(EOL);
		}
		// Jeli to nie podkonfiguracja
		else
		{
			W.WriteChar(_T(' '));
			W.WriteChar(_T('='));
			W.WriteChar(_T(' '));
			// Jeli to warto
			if (Value *SubValue = dynamic_cast<Value*>(it->second.get()))
			{
				tstring Escaped;
				Tokenizer::Escape(&Escaped, SubValue->Data, 0);
				W.WriteChar(_T('\"'));
				W.WriteString(Escaped);
				W.WriteChar(_T('\"'));
				W.WriteString(EOL);
			}
			// Jeli to lista
			else if (List *SubList = dynamic_cast<List*>(it->second.get()))
			{
				W.WriteChar(_T('{'));
				W.WriteChar(_T(' '));

				for (size_t Index = 0; Index < SubList->Data.size(); Index++)
				{
					tstring Escaped;
					Tokenizer::Escape(&Escaped, SubList->Data[Index], 0);
					W.WriteChar(_T('\"'));
					W.WriteString(Escaped);
					W.WriteChar(_T('\"'));
					if (Index < SubList->Data.size()-1)
						W.WriteChar(_T(','));
					W.WriteChar(_T(' '));
				}
				W.WriteChar(_T('}'));
				W.WriteString(EOL);
			}
			// Bd wewntrzny
			else
				assert(0 && "Impossible!");
		}
	}
}


//HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH
// Klasa Config

Config::Config() :
	pimpl(new Config_pimpl)
{
}

Config::~Config()
{
}

/*void Config::CreateError(const tstring &Msg)
{
	throw Error(Msg);
}*/

void Config::LoadFromString(const tstring &S)
{
	ERR_TRY;

	pimpl->ItemMap.clear();
	Tokenizer T(&S, Tokenizer::FLAG_MULTILINE_STRINGS);
	T.Next();
	pimpl->ReadConfig(T, false);

	ERR_CATCH(_T("Cannot read configuration from tstring."));
}

void Config::LoadFromStream(Stream *S)
{
	ERR_TRY;

	pimpl->ItemMap.clear();
	Tokenizer T(S, Tokenizer::FLAG_MULTILINE_STRINGS);
	T.Next();
	pimpl->ReadConfig(T, false);

	ERR_CATCH(_T("Cannot read configuration from stream."));
}

void Config::LoadFromFile(const tstring &FileName)
{
	ERR_TRY;

	pimpl->ItemMap.clear();
	FileStream File(FileName, FM_READ);
	Tokenizer T(&File, Tokenizer::FLAG_MULTILINE_STRINGS);
	T.Next();
	pimpl->ReadConfig(T, false);

	ERR_CATCH(_T("Cannot read configuration from file: ") + FileName);
}

void Config::SaveToString(tstring *S) const
{
	ERR_TRY;

	S->clear();

#ifdef _UNICODE
	VectorStream Stream = VectorStream();
	{
		// Tu uwaga! Nawiasy s po to eby si Writer zniszczy i na koniec zrobi swj Flush,
		// zanim wykorzystamy strumie do ktrego zapisuje.
		TCharWriter Writer(&Stream);
		pimpl->WriteConfig(Writer, 0);
	}
	S->assign((const wchar_t*)Stream.Data(), Stream.GetSize() / sizeof(wchar_t));
#else
	StringStream StringStream(S);
	TCharWriter Writer(&StringStream);
	pimpl->WriteConfig(Writer, 0);
#endif

	ERR_CATCH(_T("Cannot write configuration to tstring."));
}

void Config::SaveToStream(Stream *S) const
{
	ERR_TRY;

	TCharWriter Writer(S);
	pimpl->WriteConfig(Writer, 0);

	ERR_CATCH(_T("Cannot write configuration to stream."));
}

void Config::SaveToFile(const tstring &FileName) const
{
	ERR_TRY;

	FileStream File(FileName, FM_WRITE);
	TCharWriter Writer(&File);
	pimpl->WriteConfig(Writer, 0);

	ERR_CATCH(_T("Cannot write configuration to file: ") + FileName);
}

Item * Config::GetItem(const tstring &Path)
{
	size_t SlashPos = Path.find(_T('/'));
	// Nie ma ju w ciece nastpnego slasha - to co w tym configu
	if (SlashPos == tstring::npos)
	{
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Path);
		if (it == pimpl->ItemMap.end())
			return NULL;
		else
			return it->second.get();
	}
	// Jest nastpny slash - podaj dalej
	else
	{
		tstring Begin = Path.substr(0, SlashPos);
		tstring Rest = Path.substr(SlashPos+1);
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Begin);
		if (it == pimpl->ItemMap.end())
			return NULL;
		else
		{
			if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
				return SubConfig->GetItem(Rest);
			else
				return NULL;
		}
	}
}

Item * Config::MustGetItem(const tstring &Path)
{
	Item * R = GetItem(Path);
	if (R == NULL)
		throw Error(_T("Cannot find element with path: ") + Path);
	return R;
}

bool Config::GetData(const tstring &Path, tstring *Data)
{
	Item *item = GetItem(Path);
	if (item == NULL)
		return false;
	if (Value *di = dynamic_cast<Value*>(item))
	{
		*Data = di->Data;
		return true;
	}
	else
		return false;
}

void Config::MustGetData(const tstring &Path, tstring *Data)
{
	if (!GetData(Path, Data))
		throw Error(_T("Cannot read element data from path: ") + Path);
}

bool Config::ItemExists(const tstring &Path)
{
	size_t SlashPos = Path.find(_T('/'));
	// Nie ma ju w ciece nastpnego slasha - to co w tym configu
	if (SlashPos == tstring::npos)
		return pimpl->ItemMap.find(Path) != pimpl->ItemMap.end();
	// Jest nastpny slash - podaj dalej
	else
	{
		tstring Begin = Path.substr(0, SlashPos);
		tstring Rest = Path.substr(SlashPos+1);
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Begin);
		if (it == pimpl->ItemMap.end())
			return false;
		else
		{
			if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
				return SubConfig->ItemExists(Rest);
			else
				return false;
		}
	}
}

void Config::ItemMustExists(const tstring &Path)
{
	if (!ItemExists(Path))
		throw Error(_T("Configuration element with path: \"") + Path + _T("\" does not exist."));
}

bool Config::SetItem(const tstring &Path, Item *a_Item, uint4 CreateMode)
{
	size_t SlashPos = Path.find('/');
	// Nie ma ju w ciece nastpnego slasha - to co w tym configu
	if (SlashPos == tstring::npos)
	{
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Path);
		// Element nie istnieje
		if (it == pimpl->ItemMap.end())
		{
			// Musi istnie - bd
			if ((CreateMode & CM_MUST_EXIST))
				return false;
			// Nie musi - tworzymy
			pimpl->ItemMap.insert(std::make_pair(Path, shared_ptr<Item>(a_Item)));
			return true;
		}
		// Element istnieje
		else
		{
			// Musi nie istnie - bd
			if ((CreateMode & CM_MUST_NOT_EXIST))
				return false;
			// Musi by tego samego typu i jest innego - bd
			if ( ((CreateMode & CM_SAME_TYPE)) && (typeid(a_Item) != typeid(it->second.get())) )
				return false;
			// OK - ustawiamy (stary sam si zwolni)
			it->second.reset(a_Item);
			return true;
		}
	}
	// Jest nastpny slash - podaj dalej
	else
	{
		tstring Begin = Path.substr(0, SlashPos);
		tstring Rest = Path.substr(SlashPos+1);
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Begin);
		if (it == pimpl->ItemMap.end())
		{
			if ((CreateMode & CM_CREATE_PATH))
			{
				Config *SubConfig = new Config();
				pimpl->ItemMap.insert(std::make_pair(Begin, shared_ptr<Item>(SubConfig)));
				return SubConfig->SetItem(Rest, a_Item, CreateMode);
			}
			else
				return false;
		}
		else
		{
			if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
				return SubConfig->SetItem(Rest, a_Item, CreateMode);
			else
				return false;
		}
	}
}

void Config::MustSetItem(const tstring &Path, Item *a_Item, uint4 CreateMode)
{
	if (!SetItem(Path, a_Item, CreateMode))
		throw Error(_T("Cannot set configuration element with path: ") + Path);
}

bool Config::SetData(const tstring &Path, const tstring &Data, uint4 CreateMode)
{
	size_t SlashPos = Path.find(_T('/'));
	// Nie ma ju w ciece nastpnego slasha - to co w tym configu
	if (SlashPos == tstring::npos)
	{
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Path);
		// Element nie istnieje
		if (it == pimpl->ItemMap.end())
		{
			// Musi istnie - bd
			if ((CreateMode & CM_MUST_EXIST))
				return false;
			// Nie musi - tworzymy
			Item *i = new Value(Data);
			pimpl->ItemMap.insert(std::make_pair(Path, shared_ptr<Item>(i)));
			return true;
		}
		// Element istnieje
		else
		{
			// Musi nie istnie - bd
			if ((CreateMode & CM_MUST_NOT_EXIST))
				return false;
			// Musi by Value
			if (Value *di = dynamic_cast<Value*>(it->second.get()))
			{
				// Ustawiamy
				di->Data = Data;
				return true;
			}
			else
				return false;
		}
	}
	// Jest nastpny slash - podaj dalej
	else
	{
		tstring Begin = Path.substr(0, SlashPos);
		tstring Rest = Path.substr(SlashPos+1);
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Begin);
		if (it == pimpl->ItemMap.end())
		{
			if ((CreateMode & CM_CREATE_PATH))
			{
				Config *SubConfig = new Config();
				pimpl->ItemMap.insert(std::make_pair(Begin, shared_ptr<Item>(SubConfig)));
				return SubConfig->SetData(Rest, Data, CreateMode);
			}
			else
				return false;
		}
		else
		{
			if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
				return SubConfig->SetData(Rest, Data, CreateMode);
			else
				return false;
		}
	}
}

void Config::MustSetData(const tstring &Path, const tstring &Data, uint4 CreateMode)
{
	if (!SetData(Path, Data, CreateMode))
		throw Error(_T("Cannot write configuration data to path: ") + Path);
}

bool Config::DeleteItem(const tstring &Path)
{
	size_t SlashPos = Path.find('/');
	// Nie ma ju w ciece nastpnego slasha - to co w tym configu
	if (SlashPos == tstring::npos)
	{
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Path);
		// Element nie istnieje
		if (it == pimpl->ItemMap.end())
			return false;
		// Element istnieje
		else
		{
			// Usu go
			pimpl->ItemMap.erase(it);
			return true;
		}
	}
	// Jest nastpny slash - podaj dalej
	else
	{
		tstring Begin = Path.substr(0, SlashPos);
		tstring Rest = Path.substr(SlashPos+1);
		ITEM_MAP::iterator it = pimpl->ItemMap.find(Begin);
		// Nie ma pod-configa o nazwie Begin
		if (it == pimpl->ItemMap.end())
			return false;
		// Przeka do pod-configa o nazwie Begin
		else
		{
			// To musi by podconfig
			if (Config *SubConfig = dynamic_cast<Config*>(it->second.get()))
				return SubConfig->DeleteItem(Rest);
			else
				return false;
		}
	}
}

void Config::MustDeleteItem(const tstring &Path)
{
	if (!DeleteItem(Path))
		throw Error(_T("Cannot delete configuration element: ") + Path, __TFILE__, __LINE__);
}

void Config::GetElementNames(STRING_VECTOR *OutNames)
{
	OutNames->clear();
	ITEM_MAP::iterator it = pimpl->ItemMap.begin();
	while (it != pimpl->ItemMap.end())
	{
		OutNames->push_back(it->first);
		++it;
	}
}

scoped_ptr<Config> g_Config;

} // namespace common
