/** \file
\brief Obsuga plikw i systemu plikw
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Files \n
Elementy moduu: \ref code_files
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_FILES_H_
#define COMMON_FILES_H_

#include "Stream.hpp"


namespace common
{

struct DATETIME;

/** \addtogroup code_files Modu Files
Dokumentacja: \ref Module_Files \n
Nagwek: Files.hpp */
//@{

/// Rodzaj elementu systemu plikw
enum FILE_ITEM_TYPE
{
	IT_NONE, ///< Bd, brak, koniec itp.
	IT_DIR,  ///< Katalog
	IT_FILE, ///< Plik
};

/// Tryb otwarcia pliku
enum FILE_MODE
{
	/** zapis: tak, odczyt: nie \n
	pozycja: 0 \n
	nie istnieje: tworzy, istnieje: oprnia \n
	SetPos: dziaa */
	FM_WRITE,
	/** zapis: tak, odczyt: tak \n
	pozycja: 0 \n
	nie istnieje: tworzy, istnieje: oprnia \n
	SetPos: dziaa */
	FM_WRITE_PLUS,
	/** zapis: nie, odczyt: tak \n
	pozycja: 0 \n
	nie istnieje: bd, istnieje: otwiera \n
	SetPos: dziaa */
	FM_READ,
	/** zapis: tak, odczyt: tak \n
	pozycja: 0 \n
	nie istnieje: bd, istnieje: otwiera \n
	SetPos: dziaa */
	FM_READ_PLUS,
	/** zapis: tak, odczyt: nie \n
	pozycja: koniec \n
	nie istnieje: tworzy, istnieje: otwiera \n
	SetPos: w Linuksie nie dziaa! */
	FM_APPEND,
	/** zapis: tak, odczyt: tak \n
	pozycja: koniec \n
	nie istnieje: tworzy, istnieje: otwiera \n
	SetPos: w Linuksie nie dziaa! */
	FM_APPEND_PLUS,
};

/// \internal
class File_pimpl;

/// Strumie plikowy
class FileStream : public SeekableStream
{
private:
	scoped_ptr<File_pimpl> pimpl;

public:
	FileStream(const tstring &FileName, FILE_MODE FileMode, bool Lock = true);
	virtual ~FileStream();

	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void Flush();
	virtual size_t GetSize();
	virtual int GetPos();
	virtual void SetPos(int pos);
	virtual void SetPosFromCurrent(int pos);
	virtual void SetPosFromEnd(int pos);
	virtual void SetSize(size_t Size);
	virtual void Truncate();
	virtual bool End();
};

/// \internal
class DirLister_pimpl;

/// Klasa do listowania zawartoci katalogu.
/** W przypadku bdw rzuca wyjtki.
Nie listuje <tt>"."</tt> ani <tt>".."</tt>. */
class DirLister
{
private:
	scoped_ptr<DirLister_pimpl> pimpl;

public:
	DirLister(const tstring &Dir);
	~DirLister();

	/// Pobiera pierwszy/nastpny element (plik lub katalog) i zwraca go przez parametry.
	/** Jeli nie udao si, bo ju koniec, zawraca false.
	\param[out] OutName to sama nazwa, nie pena cieka.
	\param[out] OutType to DIR lub FILE. */
	bool ReadNext(tstring *OutName, FILE_ITEM_TYPE *OutType);
};


/// Zapisuje podany acuch jako tre pliku
void SaveStringToFile(const tstring &FileName, const string &Data);
/// Zapisuje podany acuch jako tre pliku
void SaveDataToFile(const tstring &FileName, const void *Data, size_t NumBytes);
/// Wczytuje ca zawarto pliku do acucha
void LoadStringFromFile(const tstring &FileName, string *Data);
/// Wersje Unicode
#ifdef WIN32
void SaveStringToFile(const tstring &FileName, const wstring &Data);
void LoadStringFromFile(const tstring &FileName, wstring *Data);
#endif

/// Zwraca wybrane informacje na temat pliku/katalogu
/** Jako parametry wyjciowe mona podawa NULL.
Jeli to katalog, zwrcony rozmiar jest niezdefiniowany. */
bool GetFileItemInfo(const tstring &Path, FILE_ITEM_TYPE *OutType, uint *OutSize, DATETIME *OutModificationTime, DATETIME *OutCreationTime = NULL, DATETIME *OutAccessTime = NULL);
void MustGetFileItemInfo(const tstring &Path, FILE_ITEM_TYPE *OutType, uint *OutSize, DATETIME *OutModificationTime, DATETIME *OutCreationTime = NULL, DATETIME *OutAccessTime = NULL);
/// Zwraca typ - czy to jest plik, katalog czy bd/nie-istnieje.
/** Jeli bd lub nie istnieje, zwraca common::IT_NONE. */
FILE_ITEM_TYPE GetFileItemType(const tstring &Path);

/// Ustawia czas dostpu i modyfikacji podanego pliku na biecy
bool UpdateFileTimeToNow(const tstring &FileName);
void MustUpdateFileTimeToNow(const tstring &FileName);
/// Ustawia czas dostpu i modyfikacji podanego pliku na podane
bool UpdateFileTime(const tstring &FileName, const DATETIME &ModificationTime, const DATETIME &AccessTime);
void MustUpdateFileTime(const tstring &FileName, const DATETIME &ModificationTime, const DATETIME &AccessTime);

/// Tworzy katalog
bool CreateDirectory(const tstring &Path);
void MustCreateDirectory(const tstring &Path);
/// Usuwa katalog
/** Katalog musi by pusty. */
bool DeleteDirectory(const tstring &Path);
void MustDeleteDirectory(const tstring &Path);
/// Tworzy ca ciek katalogw a po podanego, np. <tt>"C:\A\B\C"</tt> nawet jeli nie istnia nawet <tt>"C:\A"</tt>
bool CreateDirectoryChain(const tstring &Path);
void MustCreateDirectoryChain(const tstring &Path);

/// Usuwa plik
bool DeleteFile(const tstring &FileName);
void MustDeleteFile(const tstring &FileName);
/// Przenosi lub zmienia nazw pliku lub katalogu
/** Katalogi mona przenosi tylko w ramach jednej partycji. Pliki niekoniecznie. */
bool MoveItem(const tstring &OldPath, const tstring &NewPath);
void MustMoveItem(const tstring &OldPath, const tstring &NewPath);

// Pliki tekstowe z uyciem Unicode
#ifdef WIN32

/// Kodowanie pliku tekstowego
enum FILE_ENCODING
{
	/** \name Modsze 16 bitw - kodowanie */
	//@{
	FILE_ENCODING_UTF16_LE = 1, ///< Zwyky UTF-16 stosowany w Windows (domylnie dodaje BOM)
	FILE_ENCODING_UTF8     = 2, ///< UTF-8 (domylnie nie dodaje BOM)
	FILE_ENCODING_ANSI     = 3, ///< Natywne kodowanie systemu
	//@}

	/** \name Starsze 16 bitw - flagi dodatkowe */
	//@{
	/// Docz t flag dodatkowo | podczas zapisywania, aby wymusi nagwek BOM.
	FILE_ENCODING_FORCE_BOM  = 0x00010000,
	/// Docz t flag dodatkowo | podczas zapisywania, aby wymusi brak nagwka BOM.
	FILE_ENCODING_NO_BOM     = 0x00020000,
	/// Docz t flag dodatkowo podczas odczytywania, aby automatycznie wykry kodowanie.
	FILE_ENCODING_AUTODETECT = 0x00040000,
	//@}
};

/// Zapisuje tekst Unicode do pliku w podanym kodowaniu
/** Musisz poda jedn flag common::FILE_ENCODING z kodowaniem.
Moesz doda jedn z flag wymuszajc nagwek BOM lub jego brak. */
void SaveUnicodeToFile(const tstring &FileName, const wstring &Data, unsigned Encoding);
void SaveUnicodeToFile(const tstring &FileName, const wchar_t *Data, unsigned NumChars, unsigned Encoding);

void SaveUnicodeToStream(Stream *Dest, const wstring &Data, unsigned Encoding);
void SaveUnicodeToStream(Stream *Dest, const wchar_t *Data, unsigned NumChars, unsigned Encoding);

/// Wczytuje plik tekstowy do acucha Unicode
/** \param Encoding:
- Moesz poda samo kodowanie, aby wczyta w z gry zaoonym kodowaniu (duo szybsze!)
- Moesz poda samo common::FILE_ENCODING_AUTODETECT, aby automatycznie wykry kodowanie.
- Moesz poda common::FILE_ENCODING_AUTODETECT i doda jedno z kodowa, aby automatycznie wykry,
  a w przypadku wtpliwoci wybra podane.
\param OutEncoding - zwraca wykryte kodowanie. Jeli ci nie interesuj, moesz poda NULL.
*/
void LoadUnicodeFromFile(const tstring &FileName, wstring *Out, unsigned Encoding, FILE_ENCODING *OutEncoding = NULL);
/// Wczytuje plik tekstowy do acucha ANSI.
/** Wspiera te same kodowania co common::OutEncoding. */
void LoadUnicodeFromFile(const tstring &FileName, string *Out, unsigned Encoding, FILE_ENCODING *OutEncoding = NULL);

void LoadUnicodeFromStream(SeekableStream *Src, wstring *Out, unsigned Encoding, FILE_ENCODING *OutEncoding = NULL);
void LoadUnicodeFromStream(SeekableStream *Src, string *Out, unsigned Encoding, FILE_ENCODING *OutEncoding = NULL);

#endif

//@}
// code_files

} // namespace common

#endif
