/** \file
\brief Mechanizm logowania komunikatw
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Logger \n
Elementy moduu: \ref code_logger
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_LOGGER_H_
#define COMMON_LOGGER_H_

// Dziwaczna deklaracja zapowiadajca, eby nie wcza tu w nagwku <iostream>
namespace std
{
	template <typename T> struct char_traits;
	template <typename T1, typename T2> class basic_ostream;
	typedef basic_ostream<char, char_traits<char> > ostream;
};

namespace common
{

/** \addtogroup code_logger Modu Logger
Dokumentacja: \ref Module_Logger \n
Nagwek: Logger.hpp */
//@{

/// \internal
class ILog;
/// \internal
class Logger_pimpl;

/// Logger - klasa gwna systemu logujcego.
class Logger
{
	friend void CreateLogger(bool);
	friend void DestroyLogger();
	friend class ILog;

private:
	scoped_ptr<Logger_pimpl> pimpl;

	Logger(bool UseQueue);
	~Logger();

public:
	/** \name Konfiguracja
	Te metody nie s bezpieczne wtkowo. */
	//@{
	/// Rejestruje log w loggerze mapujc do niego okrelone typy komunikatw.
	void AddLogMapping(uint4 Mask, ILog *Log);
	/// Ustawia mapowanie prefiksw typu dla wszystkich zarejestrowanych logw.
	void AddTypePrefixMapping(uint4 Mask, const tstring &Prefix);
	/// Ustawia format prefiksu dla wszystkich zarejestrowanych logw.
	void SetPrefixFormat(const tstring &PrefixFormat);
	//@}

	/** \name Uywanie loggera.
	Te metody s bezpieczne wtkowo. */
	//@{
	/// Ustawia wasn informacj prefiksu. Indeks: 0..2
	void SetCustomPrefixInfo(int Index, const tstring &Info);
	//// Loguje komunikat - najwaniejsza funkcja!
	void Log(uint4 Type, const tstring &Message);
	//@}
};

/// Abstrakcyjna kasa bazowa wszelkich logw.
class ILog
{
	friend class Logger_pimpl;

private:
	/// \internal
	class ILog_pimpl;
	scoped_ptr<ILog_pimpl> pimpl;

protected:
	/// Ma zalogowa podany komunikat tam gdzie trzeba
	virtual void OnLog(uint4 Type, const tstring &Prefix, const tstring &TypePrefix, const tstring &Message) = 0;

public:
	ILog();
	virtual ~ILog();

	/** \name Konfiguracja */
	//@{
	/// Dodaje mapowanie prefiksu typu dla tego loga
	void AddTypePrefixMapping(uint4 Mask, const tstring &Prefix);
	/// Ustawia format prefiksu dla tego loga
	void SetPrefixFormat(const tstring &PrefixFormat);
	//@}
};

/// Tworzy logger
void CreateLogger(bool UseQueue);
/// Usuwa logger
void DestroyLogger();
/// Pobiera logger
Logger & GetLogger();
/// Zwraca true, jeli Logger jest zainicjowany
bool IsLogger();


//HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH
// Konkretne loggery

/// Tryby, w jakich moe pracowa logger plikowy
enum LOG_FILE_MODE
{
	/// Normalny - najszybszy
	FILE_MODE_NORMAL,
	/// Po kadym komunikacie FLUSH
	FILE_MODE_FLUSH,
	/// Dla kadego komunikatu osobne otwarcie pliku - najbardziej niezawodny
	FILE_MODE_REOPEN,
};

/// Log do pliku TXT
class TextFileLog : public ILog
{
private:
	/// \internal
	class TextFileLog_pimpl;
	scoped_ptr<TextFileLog_pimpl> pimpl;

protected:
	virtual void OnLog(uint4 Type, const tstring &Prefix, const tstring &TypePrefix, const tstring &Message);

public:
	TextFileLog(const tstring &FileName, LOG_FILE_MODE Mode, EOLMODE EolMode, bool Append = false, const tstring &StartText = _T(""));
	virtual ~TextFileLog();
};

/// Log do pliku HTML
class HtmlFileLog : public ILog
{
public:
	struct STYLE
	{
		uint4 BackgroundColor;
		uint4 FontColor;
		bool Bold;
		bool Italic;

		STYLE(bool Bold = false, bool Italic = false, uint FontColor = 0x000000, uint BackgroundColor = 0xFFFFFF);
	};

private:
	/// \internal
	class HtmlFileLog_pimpl;
	scoped_ptr<HtmlFileLog_pimpl> pimpl;

protected:
	virtual void OnLog(uint4 Type, const tstring &Prefix, const tstring &TypePrefix, const tstring &Message);

public:
	HtmlFileLog(const tstring &FileName, LOG_FILE_MODE Mode, bool Append = false, const tstring &StartText = _T(""));
	virtual ~HtmlFileLog();

	/** \name Konfiguracja */
	//@{
	/// Dodaje mapowanie typu komunikatu na styl
	void AddStyleMapping(uint4 Mask, const STYLE &Style);
	//@}
};

/// Log zapisujcy do strumienia wyjciowego biblioteki standardowej C++ - std::ostream.
class OstreamLog : public ILog
{
private:
	class Pimpl;
	scoped_ptr<Pimpl> pimpl;

protected:
	virtual void OnLog(uint4 Type, const tstring &Prefix, const tstring &TypePrefix, const tstring &Message);

public:
#ifdef _UNICODE
	OstreamLog(std::wostream *Ostream);
#else
	OstreamLog(std::ostream *Ostream);
#endif
	virtual ~OstreamLog();
};

//@}
// code_logger

} // namespace common

/** \addtogroup code_logger */
//@{
/// Skrt do atwego zalogowania acucha
#define LOG(Type, s) { if (common::IsLogger()) common::GetLogger().Log((Type), (s)); else assert(0 && "LOG macro: Logger not initialized."); }
//@}

#endif
