/** \file
\brief Hierarchia klas strumieni
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Stream \n
Elementy moduu: \ref code_stream
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_STREAM_H_
#define COMMON_STREAM_H_

namespace common
{

/** \addtogroup code_stream Modu Stream
Dokumentacja: \ref Module_Stream \n
Nagwek: Stream.hpp */
//@{

extern const size_t BUFFER_SIZE;

/// \internal
void _ThrowBufEndError(const tchar *File, int Line);

/// Tolerancja bdw podczas dekodowania danych zapisanych tekstowo
enum DECODE_TOLERANCE
{
	DECODE_TOLERANCE_NONE,       ///< adne dodatkowe znaki nie bd tolerowane
	DECODE_TOLERANCE_WHITESPACE, ///< Biae znaki bd ignorowane i nie spowoduj bdu
	DECODE_TOLERANCE_ALL,        ///< Wszelkie nieznane znaki bd ignorowane i nie spowoduj bdu
};


/// Abstrakcyjna klasa bazowa strumieni danych binarnych
class Stream
{
public:
	virtual ~Stream() { }

	/** \name Zapisywanie */
	//@{

	/// Zapisuje dane
	/** (W oryginale: zgasza bd) */
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush() { }

	/// Zapisuje dane, sama odczytuje rozmiar przekazanej zmiennej
	template <typename T>
	void WriteEx(const T &x) { return Write(&x, sizeof(x)); }
	/// Zapisuje dugo na 1B i acuch
	/** Jeli acuch za dugi - bd. */
	void WriteString1(const string &s);
	void WriteString1(const char *s);
	void WriteString1(const char *s, unsigned NumChars);
#ifdef WIN32
	void WriteString1(const wstring &s);
	void WriteString1(const wchar_t *s);
	void WriteString1(const wchar_t *s, unsigned NumChars);
#endif
	/// Zapisuje dugo na 2B i acuch
	/** Jeli acuch za dugi - bd. */
	void WriteString2(const string &s);
	void WriteString2(const char *s);
	void WriteString2(const char *s, unsigned NumChars);
#ifdef WIN32
	void WriteString2(const wstring &s);
	void WriteString2(const wchar_t *s);
	void WriteString2(const wchar_t *s, unsigned NumChars);
#endif
	/// Zapisuje dugo na 4B i acuch
	void WriteString4(const string &s);
	void WriteString4(const char *s);
	void WriteString4(const char *s, unsigned NumChars);
#ifdef WIN32
	void WriteString4(const wstring &s);
	void WriteString4(const wchar_t *s);
	void WriteString4(const wchar_t *s, unsigned NumChars);
#endif
	/// Zapisuje acuch bez dugoci
	void WriteStringF(const string &s);
#ifdef WIN32
	void WriteStringF(const wstring &s);
#endif
	/// Zapisuje warto logiczn za pomoc jednego bajtu
	void WriteBool(bool b);

	//@}

	/** \name Odczytywanie */
	//@{

	/// Odczytuje dane
	/** Jeli osignito koniec, funkcja nie zwraca bdu, a liczba odczytanych bajtw to 0.
	(W oryginale: zgasza bd)
	\param Size Liczba bajtw do odczytania
	\return Zwraca liczb bajtw, jakie udao si odczyta
	*/
	virtual size_t Read(void *Data, size_t MaxLength);
	/// Odczytuje tyle bajtw, ile si zada
	/** Jeli koniec pliku albo jeli odczytano mniej, zgasza bd.
	(Mona j przeadowa, ale nie trzeba - ma swoj wersj oryginaln)
	*/
	virtual void MustRead(void *Data, size_t Length);
	/// Tak samo jak MustRead(), ale sama odczytuje rozmiar przekazanej zmiennej
	/** Zwraca true, jeli osignito koniec strumienia
	(W oryginale: zgasza bd)
	*/
	virtual bool End();
	/// Pomija co najwyej podan liczb bajtw (chyba e wczeniej osignito koniec).
	/** (Mona j przeadowa, ale nie trzeba - ma swoj wersj oryginaln)
	\return Zwraca liczb pominitych bajtw.
	*/
	virtual size_t Skip(size_t MaxLength);

	/// Odczytuje dane, sama odczytuje rozmiar przekazanej zmiennej
	template <typename T>
	void ReadEx(T *x) { MustRead(x, sizeof(*x)); }
	/// Odczytuje acuch poprzedzony rozmiarem 1B
	void ReadString1(string *s);
	/// Odczytuje acuch poprzedzony rozmiarem 2B
	void ReadString2(string *s);
	/// Odczytuje acuch poprzedzony rozmiarem 4B
	void ReadString4(string *s);
	/// Odczytuje acuch bez zapisanego rozmiaru
	void ReadStringF(string *s, size_t NumChars);
	/// Odczytuje tyle znakw do acucha, ile si da do koca strumienia
	void ReadStringToEnd(string *s);
	/// Wersje powyszych funkcji dla Unicode
#ifdef WIN32
	void ReadString1(wstring *s);
	void ReadString2(wstring *s);
	void ReadString4(wstring *s);
	void ReadStringF(wstring *s, size_t NumChars);
	void ReadStringToEnd(wstring *s);
#endif
	/// Odczytuje warto logiczn za pomoc jednego bajtu
	void ReadBool(bool *b);
	/// Pomija koniecznie podan liczb bajtw.
	/** Jeli wczeniej osignito koniec, zgasza bd. */
	void MustSkip(size_t Length);

	//@}

	/** \name Kopiowanie */
	//@{
	/// Odczytuje podan co najwyej liczb bajtw z podanego strumienia
	/** \return Zwraca liczb odczytanych bajtw */
	size_t CopyFrom(Stream *s, size_t Size);
	/// Odczytuje dokadnie podan liczb bajtw z podanego strumienia
	/** Jeli odczyta mniej, zgasza bd. */
	void MustCopyFrom(Stream *s, size_t Size);
	/// Odczytuje dane do koca z podanego strumienia
	void CopyFromToEnd(Stream *s);
	//@}
};

/// Abstrakcyjna klasa bazowa strumieni pozwalajcych na odczytywanie rozmiaru i zmian pozycji
class SeekableStream : public Stream
{
public:
	// ======== Implementacja Stream ========
	/** W Stream domylnie rzuca bd, tu w SeekableStream dostaje ju wersj oryginaln. Mona j dalej nadpisa. */
	virtual bool End();
	/** W Stream ma wersj oryginaln, tu w SeekableStream dostaje lepsz. Mona j dalej nadpisa. */
	virtual size_t Skip(size_t MaxLength);

	/// Zwraca rozmiar
	/** W oryginale: zgasza bd. */
	virtual size_t GetSize();
	/// Zwraca pozycj wzgldem pocztku
	/** W oryginale: zgasza bd. */
	virtual int GetPos();
	/// Ustawia pozycj wzgldem pocztku
	/** W oryginale: zgasza bd.
	\param pos Musi by dodatnie. */
	virtual void SetPos(int pos);
	/// Ustawia pozycj wzgldem biecej
	/** Mona j nadpisa, ale ma wersj oryginaln
	\param pos moe by dodatnie albo ujemne. */
	virtual void SetPosFromCurrent(int pos);
	/// Ustawia pozycj wzgldem kocowej
	/** Mona j nadpisa, ale ma wersj oryginaln.
	Np. abu ustawi na ostatni znak, przesu do -1. */
	virtual void SetPosFromEnd(int pos);
	/// Ustawia pozycj na pocztku
	/** Mona j nadpisa, ale ma wersj oryginaln. */
	virtual void Rewind();
	/// Ustawia rozmiar na podany
	/** Po tym wywoaniu pozycja kursora jest niezdefiniowana.
	W oryginale: zgasza bd. */
	virtual void SetSize(size_t Size);
	/// Obcina strumie od biecej pozycji do koca, czyli ustawia taki rozmiar jak bieca pozycja.
	/** Mona j nadpisa, ale ma wersj oryginaln. */
	virtual void Truncate();
	/// Oprnia strumie
	/** Mona j nadpisa, ale ma wersj oryginaln. */
	virtual void Clear();
};

/// Nakadka przyspieszajca zapisujca do strumienia pojedyncze znaki i inne rzeczy.
/**
- Nie jest strumieniem.
- Tylko zapis.
- Jest szybka, poniewa:
  -# Nie ma metod wirtualnych.
  -# Buforuje, wic nie uywa za kadym razem metod wirtualnych strumienia.
- Sama zapisuje dane pozostae w buforze w destruktorze.
  Jeli chcesz to zrobi wczeniej i z kontrol bdw, wywoaj CharWriter::Flush().
- W trybie Debug dziaa nawet troch wolniej ni zapisywanie po bajcie prosto do strumienia.
  W trybie Release daje przyspieszenie prawie 10x wzgldem zapisywania po bajcie prosto do strumienia.
- Celowo nie uywam w tej klasie ERR_TRY .. ERR_CATCH_FUNC, eby byo szybciej.
*/
class CharWriter
{
private:
	Stream *m_Stream;
	std::vector<char> m_Buf;
	uint m_BufIndex;

	// Wykonuje Flush danych z bufora do strumienia nie sprawdzajc ju, czy bufor nie jest pusty.
	// Oczywicie zeruje te BufIndex.
	void DoFlush() { m_Stream->Write(&m_Buf[0], m_BufIndex); m_BufIndex = 0; }

public:
	CharWriter(Stream *a_Stream) : m_Stream(a_Stream), m_BufIndex(0) { m_Buf.resize(BUFFER_SIZE); }
	~CharWriter();

	/// Zapisuje pojedynczy znak
	void WriteChar(char Ch) { if (m_BufIndex == BUFFER_SIZE) DoFlush(); m_Buf[m_BufIndex++] = Ch; }
	/// Zapisuje po prostu znaki acucha, bez adnej dugoci
	void WriteString(const string &s);
	/// Zapisuje acuch zakoczony zerem
	void WriteString(const char *s) { WriteString(s, strlen(s)); }
	/// Zapisuje acuch podanej dugoci
	void WriteString(const char *s, size_t NumChars) { WriteData(s, NumChars); }
	/// Zapisuje surowe dane binarne
	void WriteData(const void *Data, size_t Length);
	/// Zapisuje dowoln informacj przerobion na posta tekstow, o ile da si przerobi
	template <typename T> void WriteEx(const T &x) { string s; SthToStr<T>(&s, x); WriteString(s); }
	/// Wymusza oprnienie bufora i wysanie pozostaych w nim danych do strumienia
	void Flush() { if (m_BufIndex > 0) DoFlush(); }
};

#ifdef WIN32
/// Nakadka przyspieszajca zapisujca do strumienia pojedyncze znaki Unicode.
/** Uwaga! Wersja Unicode nie posiada zapisywania surowych danych binarnych. */
class WCharWriter
{
private:
	Stream *m_Stream;
	std::vector<wchar_t> m_Buf;
	uint m_BufIndex;

	// Wykonuje Flush danych z bufora do strumienia nie sprawdzajc ju, czy bufor nie jest pusty.
	// Oczywicie zeruje te BufIndex.
	void DoFlush() { m_Stream->Write(&m_Buf[0], m_BufIndex * sizeof(wchar_t)); m_BufIndex = 0; }

public:
	WCharWriter(Stream *a_Stream) : m_Stream(a_Stream), m_BufIndex(0) { m_Buf.resize(BUFFER_SIZE); }
	~WCharWriter();

	/// Zapisuje pojedynczy znak
	void WriteChar(wchar_t Ch) { if (m_BufIndex == BUFFER_SIZE) DoFlush(); m_Buf[m_BufIndex++] = Ch; }
	/// Zapisuje po prostu znaki acucha, bez adnej dugoci
	void WriteString(const wstring &s);
	/// Zapisuje acuch zakoczony zerem
	void WriteString(const wchar_t *s) { WriteString(s, wcslen(s)); }
	/// Zapisuje acuch podanej dugoci
	void WriteString(const wchar_t *s, size_t NumChars);
	/// Wymusza oprnienie bufora i wysanie pozostaych w nim danych do strumienia
	void Flush() { if (m_BufIndex > 0) DoFlush(); }
};
#endif

/// Nakadka przyspieszajca odczytujca ze strumienia pojedyncze znaki i inne rzeczy.
/**
- Nie jest strumieniem.
- Tylko odczyt.
- Jest szybka, poniewa:
  -# Nie ma metod wirtualnych.
  -# Buforuje, wic nie uywa za kadym razem metod wirtualnych strumienia.
- Szybko wzgldem wczytywania pojedynczych znakw prosto ze strumienia: Debug = 10 razy, Release = 450 razy!
- Celowo nie uywam w tej klasie ERR_TRY .. ERR_CATCH_FUNC, eby byo szybciej.
*/
class CharReader
{
private:
	Stream *m_Stream;
	std::vector<char> m_Buf;
	// Miejsce, do ktrego doczytaem z bufora
	uint m_BufBeg;
	// Miejsce, do ktrego bufor jest wypeniony danymi
	uint m_BufEnd;

	// Wczytuje now porcj danych do strumienia
	// Wywoywa tylko kiedy bufor si skoczy, tzn. m_BufBeg == m_BufEnd.
	// Jeli skoczy si strumie i nic nie udao si wczyta, zwraca false.
	// Jeli nie, zapewnia e w buforze bdzie co najmniej jeden znak.
	bool EnsureNewChars();

public:
	CharReader(Stream *a_Stream) : m_Stream(a_Stream), m_BufBeg(0), m_BufEnd(0) { m_Buf.resize(BUFFER_SIZE); }

	/// Czy jestemy na kocu danych?
	bool End() {
		return (m_BufBeg == m_BufEnd) && m_Stream->End(); // Nic nie zostao w buforze i nic nie zostao w strumieniu.
	}
	/// Jeli mona odczyta nastpny znak, wczytuje go i zwraca true.
	/** Jeli nie, nie zmienia Out i zwraca false. Oznacza to koniec strumienia. */
	bool ReadChar(char *Out) { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } *Out = m_Buf[m_BufBeg++]; return true; }
	/// Jeli mona odczyta nastpny znak, wczytuje go.
	/** Jeli nie, rzuca wyjtek. */
	char MustReadChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } return m_Buf[m_BufBeg++]; }
	/// Jeli mona odczyta nastpny znak, podglda go zwracajc przez Out i zwraca true. Nie przesuwa "kursora".
	/** Jeli nie, nie zmienia Out i zwraca false. Oznacza to koniec strumienia. */
	bool PeekChar(char *Out) { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } *Out = m_Buf[m_BufBeg]; return true; }
	/// Jeli mona odczyta nastpny znak, podglda go zwracajc. Nie przesuwa "kursora".
	/** Jeli nie, rzuca wyjtek. */
	char MustPeekChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } return m_Buf[m_BufBeg]; }
	/// Wczytuje co najwyej MaxLength znakw do podanego stringa.
	/** StringStream jest czyszczony - nie musi by pusty ani zaalokowany. (???)
	\return Zwraca liczb odczytanych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t ReadString(string *Out, size_t MaxLength);
	/// Wczytuje Length znakw do podanego stringa.
	/** StringStream jest czyszczony - nie musi by pusty ani zaalokowany. (???)
	Jeli nie uda si odczyta wszystkich Length znakw, rzuca wyjtek. */
	void MustReadString(string *Out, size_t Length);
	/// Wczytuje co najwyej MaxLength znakw pod podany wskanik.
	/** \param[out] Out musi by tablic zaalokowan do dugoci co najmniej MaxLength.
	\return Zwraca liczb odczytanych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t ReadString(char *Out, size_t MaxLength);
	/// Wczytuje Length znakw pod podany wskanik.
	/** Jeli nie uda si odczyta wszystkich Length znakw, rzuca wyjtek.
	\param[out] Out musi by tablic zaalokowan do dugoci co najmniej Length. */
	void MustReadString(char *Out, size_t Length);
	/// Wczytuje co najwyej MaxLength bajtw pod podany wskanik.
	/** \param[out] Out musi by tablic zaalokowan do dugoci co najmniej MaxLength bajtw.
	\return Zwraca liczb odczytanych bajtw. Mniej ni dano oznacza koniec strumienia. */
	size_t ReadData(void *Out, size_t MaxLength);
	/// Wczytuje Length bajtw pod podany wskanik.
	/** Jeli nie uda si odczyta wszystkich Length bajtw, rzuca wyjtek.
	\param[out] Out musi by tablic zaalokowan do dugoci co najmniej Length bajtw. */
	void MustReadData(void *Out, size_t Length);
	/// Jeli jest nastpny znak do odczytania, pomija go i zwraca true.
	/** Jeli nie, zwraca false. Oznacza to koniec strumienia. */
	bool SkipChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } m_BufBeg++; return true; }
	/// Jeli jest nastpny znak do odczytania, pomija go.
	/** Jeli nie, rzuca wyjtek. */
	void MustSkipChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } m_BufBeg++; }
	/// Pomija co najwyej MaxLength znakw.
	/** \return Zwraca liczb pominitych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t Skip(size_t MaxLength);
	/// Pomija Length znakw.
	/** Jeli nie uda si pomin wszystkich Length znakw, rzuca wyjtek. */
	void MustSkip(size_t Length);
	/// Wczytuje linijk tekstu i jeli si udao, zwraca true.
	/** Jeli nie udao si wczyta ani jednego znaku, bo to ju koniec strumienia, zwraca false.
	Czyta znaki do napotkania koca strumienia lub koca wiersza, ktrym s <tt>"\r"</tt>, <tt>"\n"</tt> i <tt>"\r\n"</tt>.
	Koniec wiersza przeskakuje, ale nie docza go do Out. */
	bool ReadLine(string *Out);
	/// Wczytuje linijk tekstu.
	/** Jeli nie udao si wczyta ani jednego znaku, bo to ju koniec strumienia, rzuca wyjtek.
	Czyta znaki do napotkania koca strumienia lub koca wiersza, ktrym s <tt>"\r"</tt>, <tt>"\n"</tt> i <tt>"\r\n"</tt>.
	Koniec wiersza przeskakuje, ale nie docza go do Out. */
	void MustReadLine(string *Out);
};

#ifdef WIN32
/// Nakadka przyspieszajca odczytujca ze strumienia pojedyncze znaki Unicode.
/** Uwaga! Wersja Unicode nie posiada odczytywania surowych danych binarnych. */
class WCharReader
{
private:
	Stream *m_Stream;
	std::vector<wchar_t> m_Buf;
	// Miejsce, do ktrego doczytaem z bufora
	uint m_BufBeg;
	// Miejsce, do ktrego bufor jest wypeniony danymi
	uint m_BufEnd;

	// Wczytuje now porcj danych do strumienia
	// Wywoywa tylko kiedy bufor si skoczy, tzn. m_BufBeg == m_BufEnd.
	// Jeli skoczy si strumie i nic nie udao si wczyta, zwraca false.
	// Jeli nie, zapewnia e w buforze bdzie co najmniej jeden znak.
	bool EnsureNewChars();

public:
	WCharReader(Stream *a_Stream) : m_Stream(a_Stream), m_BufBeg(0), m_BufEnd(0) { m_Buf.resize(BUFFER_SIZE); }

	/// Czy jestemy na kocu danych?
	bool End() {
		return (m_BufBeg == m_BufEnd) && m_Stream->End(); // Nic nie zostao w buforze i nic nie zostao w strumieniu.
	}
	/// Jeli mona odczyta nastpny znak, wczytuje go i zwraca true.
	/** Jeli nie, nie zmienia Out i zwraca false. Oznacza to koniec strumienia. */
	bool ReadChar(wchar_t *Out) { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } *Out = m_Buf[m_BufBeg++]; return true; }
	/// Jeli mona odczyta nastpny znak, wczytuje go.
	/** Jeli nie, rzuca wyjtek. */
	wchar_t MustReadChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } return m_Buf[m_BufBeg++]; }
	/// Jeli mona odczyta nastpny znak, podglda go zwracajc przez Out i zwraca true. Nie przesuwa "kursora".
	/** Jeli nie, nie zmienia Out i zwraca false. Oznacza to koniec strumienia. */
	bool PeekChar(wchar_t *Out) { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } *Out = m_Buf[m_BufBeg]; return true; }
	/// Jeli mona odczyta nastpny znak, podglda go zwracajc. Nie przesuwa "kursora".
	/** Jeli nie, rzuca wyjtek. */
	wchar_t MustPeekChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } return m_Buf[m_BufBeg]; }
	/// Wczytuje co najwyej MaxLength znakw do podanego stringa.
	/** StringStream jest czyszczony - nie musi by pusty ani zaalokowany. (???)
	\return Zwraca liczb odczytanych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t ReadString(wstring *Out, size_t MaxLength);
	/// Wczytuje Length znakw do podanego stringa.
	/** StringStream jest czyszczony - nie musi by pusty ani zaalokowany. (???)
	Jeli nie uda si odczyta wszystkich Length znakw, rzuca wyjtek. */
	void MustReadString(wstring *Out, size_t Length);
	/// Wczytuje co najwyej MaxLength znakw pod podany wskanik.
	/** \param[out] Out musi by tablic zaalokowan do dugoci co najmniej MaxLength.
	\return Zwraca liczb odczytanych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t ReadString(wchar_t *Out, size_t MaxLength);
	/// Wczytuje Length znakw pod podany wskanik.
	/** Jeli nie uda si odczyta wszystkich Length znakw, rzuca wyjtek.
	\param[out] Out musi by tablic zaalokowan do dugoci co najmniej Length. */
	void MustReadString(wchar_t *Out, size_t Length);
	/// Jeli jest nastpny znak do odczytania, pomija go i zwraca true.
	/** Jeli nie, zwraca false. Oznacza to koniec strumienia. */
	bool SkipChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) return false; } m_BufBeg++; return true; }
	/// Jeli jest nastpny znak do odczytania, pomija go.
	/** Jeli nie, rzuca wyjtek. */
	void MustSkipChar() { if (m_BufBeg == m_BufEnd) { if (!EnsureNewChars()) _ThrowBufEndError(__TFILE__, __LINE__); } m_BufBeg++; }
	/// Pomija co najwyej MaxLength znakw.
	/** \return Zwraca liczb pominitych znakw. Mniej ni dano oznacza koniec strumienia. */
	size_t Skip(size_t MaxLength);
	/// Pomija Length znakw.
	/** Jeli nie uda si pomin wszystkich Length znakw, rzuca wyjtek. */
	void MustSkip(size_t Length);
	/// Wczytuje linijk tekstu i jeli si udao, zwraca true.
	/** Jeli nie udao si wczyta ani jednego znaku, bo to ju koniec strumienia, zwraca false.
	Czyta znaki do napotkania koca strumienia lub koca wiersza, ktrym s <tt>"\r"</tt>, <tt>"\n"</tt> i <tt>"\r\n"</tt>.
	Koniec wiersza przeskakuje, ale nie docza go do Out. */
	bool ReadLine(wstring *Out);
	/// Wczytuje linijk tekstu.
	/** Jeli nie udao si wczyta ani jednego znaku, bo to ju koniec strumienia, rzuca wyjtek.
	Czyta znaki do napotkania koca strumienia lub koca wiersza, ktrym s <tt>"\r"</tt>, <tt>"\n"</tt> i <tt>"\r\n"</tt>.
	Koniec wiersza przeskakuje, ale nie docza go do Out. */
	void MustReadLine(wstring *Out);
};
#endif

#ifdef _UNICODE
	#define TCharWriter WCharWriter
	#define TCharReader WCharReader
#else
	#define TCharWriter CharWriter
	#define TCharReader CharReader
#endif

/// Strumie pamici statycznej
/** Strumie ma sta dugo i nie moe by rozszerzany - ani automatycznie, ani rcznie. */
class MemoryStream : public SeekableStream
{
private:
	char *m_Data;
	size_t m_Size;
	int m_Pos;
	bool m_InternalAlloc;

public:
	/** \param Data Podaj NULL, aby strumie sam zaalkowa, a w destruktorze zwolni pami.
	Podaj wskanik, jeli ma korzysta z twojego obszaru pamici.
	*/
	MemoryStream(size_t Size, void *Data = 0);
	virtual ~MemoryStream();

	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void MustRead(void *Data, size_t Size);

	virtual size_t GetSize();
	virtual int GetPos();
	virtual void SetPos(int pos);
	virtual void Rewind();

	char *Data() { return m_Data; }
};

/// Strumie pamici, ale dynamicznej.
/** Strumie mona rozszerza i sam te si rozszerza. */
class VectorStream : public SeekableStream
{
private:
	char *m_Data;
	size_t m_Size;
	size_t m_Capacity;
	int m_Pos;

	// Zmienia capacity, a tym samym przealokowuje pami.
	// O niczym sama nie decyduje ani niczego nie sprawdza.
	void Reserve(size_t NewCapacity);

public:
	VectorStream();
	virtual ~VectorStream();

	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void MustRead(void *Data, size_t Size);
	virtual size_t GetSize() { return m_Size; }
	virtual int GetPos() { return m_Pos; }
	virtual void SetPos(int pos) { m_Pos = pos; }
	virtual void Rewind() { m_Pos = 0; }
	virtual void SetSize(size_t Size);

	size_t GetCapacity() { return m_Capacity; }
	void SetCapacity(size_t Capacity);
	/** Nie musi pozosta aktualny po wywoaniach metod tej klasy!
	(moe zosta przealokowany) */
	char *Data() { return m_Data; }
};

/// Strumie oparty na acuchu std::string
/** Potrafi si rozszerza. */
class StringStream : public SeekableStream
{
private:
	std::string *m_Data;
	int m_Pos;
	bool m_InternalAlloc;

public:
	/** \param Data Podaj NULL, aby strumie sam zaalokowa, a w destruktorze zwolni acuch.
	Podaj wskanik, jeli ma korzysta z twojego obszaru pamici. */
	StringStream(string *Data = 0);
	virtual ~StringStream();

	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void MustRead(void *Data, size_t Size);
	virtual size_t GetSize() { return m_Data->length(); }
	virtual int GetPos() { return m_Pos; }
	virtual void SetPos(int pos) { m_Pos = pos; }
	virtual void Rewind() { m_Pos = 0; }
	virtual void SetSize(size_t Size) { m_Data->resize(Size); }
	virtual void Clear() { m_Data->clear(); }

	size_t GetCapacity() { return m_Data->capacity(); }
	void SetCapacity(size_t Capacity) { m_Data->reserve(Capacity); }
	string *Data() { return m_Data; }
};

/// Abstrakcyjna klasa bazowa nakadek na strumienie
class OverlayStream : public Stream
{
private:
	Stream *m_Stream;

public:
	OverlayStream(Stream *a_Stream) : m_Stream(a_Stream) { }

	Stream * GetStream() { return m_Stream; }

	// ======== Implementacja Stream ========
	virtual bool End() { return m_Stream->End(); }
	virtual void Flush() { m_Stream->Flush(); }
};

/// Nakadka na strumie zliczajca ilo zapisanych i odczytanych bajtw
class CounterOverlayStream : public OverlayStream
{
private:
	uint4 m_WriteCounter;
	uint4 m_ReadCounter;

public:
	CounterOverlayStream(Stream *a_Stream);

	// ====== Implementacja Stream ======
	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void MustRead(void *Data, size_t Size);
	virtual void Flush();

	uint4 GetWriteCounter() { return m_WriteCounter; }
	uint4 GetReadCounter() { return m_ReadCounter; }
	void ResetWriteCounter() { m_WriteCounter = 0; }
	void ResetReadCounter() { m_ReadCounter = 0; }
};

/// Nakadka na strumie ograniczajca liczb odczytywanych i zapisywanych bajtw.
/** Jesli przekroczy limit podczas zapisywania, zapisuje ile si da i rzuca bd zapisu.
Jeli przekroczy limit podczas odczytywania zachowuje si tak, jakby strumie osign koniec. */
class LimitOverlayStream : public OverlayStream
{
private:
	// Przechowuj aktualny limit, jaki pozosta.
	// S zmniejszane przez funkcje odczytujce i zapisujce.
	uint4 m_WriteLimit;
	uint4 m_ReadLimit;

public:
	LimitOverlayStream(Stream *a_Stream, uint4 WriteLimit, uint4 ReadLimit);

	// ====== Implementacja Stream ======
	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);
	virtual void Flush();

	/// Ustawia nowy limit
	void SetWriteLimit(uint4 WriteLimit);
	/// Ustawia nowy limit
	void SetReadLimit(uint4 ReadLimit);
};

/// Zapisuje zapisywane dane do wielu podczonych do niego strumieni na raz.
class MultiWriterStream : public Stream
{
private:
	std::vector<Stream*> m_Streams;

public:
	MultiWriterStream() { }
	MultiWriterStream(Stream *a_Streams[], uint StreamCount);

	/// Dodaje strumie do listy.
	void AddStream(Stream *a_Stream) { m_Streams.push_back(a_Stream); }
	/// Usuwa strumie z listy. Jeli takiego nie znajdzie, nic nie robi.
	void RemoveStream(Stream *a_Stream);
	/// Usuwa wszystkie strumienie z listy.
	void ClearStreams() { m_Streams.clear(); }
	size_t GetStreamCount() { return m_Streams.size(); }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush();
};

/// Klasa obliczajca hash 32-bitowy z kolejno podawanych blokw danych
/** Algorytm: Jenkins One-at-a-time hash, http://www.burtleburtle.net/bob/hash/doobs.html
Strumie tylko do zapisu.
Prawidowe uycie: Hash_Calc::Write(), Hash_Calc::Write() (lub inne funkcje zapisujce), ..., Hash_Calc::Finish().
Po wywoaniu Hash_Calc::Finish() nie mona dalej zapisywa danych!
Hash_Calc::Reset() - rozpoczyna liczenie sumy od nowa. */
class Hash_Calc : public Stream
{
private:
	uint4 m_Hash;

public:
	Hash_Calc() : m_Hash(0) { }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);

	/// Koczy obliczenia i zwraca policzon sum
	uint4 Finish();
	/// Rozpoczyna liczenie nowej sumy
	void Reset() { m_Hash = 0; }

	// ======== Statyczne ========
	/// Po prostu oblicza sum kontroln z podanych danych
	static uint4 Calc(const void *Buf, uint4 BufLen);
	static uint4 Calc(const string &s);
#ifdef WIN32
	static uint4 Calc(const wstring &s);
#endif
};

/// Klasa obliczajca sum CRC32 z kolejno podawanych blokw danych
/** Strumie tylko do zapisu.
CRC32_Calc::GetResult() - wyliczon dotychczas sum mona otrzymywa w kadej chwili, a potem dalej dodawa nowe dane.
CRC32_Calc::Reset() - rozpoczyna liczenie nowej sumy kontrolnej. */
class CRC32_Calc : public Stream
{
private:
	uint m_CRC;

public:
	CRC32_Calc() { m_CRC = 0xFFFFFFFF; }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);

	/// Zwraca policzon dotychczas sum
	uint GetResult() { return ~m_CRC; }
	/// Rozpoczyna liczenie nowej sumy
	void Reset() { m_CRC = 0xFFFFFFFF; }

	// ======== Statyczne ========
	/// Po prostu oblicza sum kontroln z podanych danych
	static uint Calc(const void *Data, size_t DataLength);
};

/// Suma MD5
/** Suma ma 128 bitw - 16 bajtw. Jest typu uint1[16]. */
struct MD5_SUM
{
	uint1 Data[16];

	uint1 & operator [] (size_t i) { return Data[i]; }
	uint1 operator [] (size_t i) const { return Data[i]; }

	bool operator == (const MD5_SUM &s) const
	{
		return
			Data[ 0] == s.Data[ 0] && Data[ 1] == s.Data[ 1] && Data[ 2] == s.Data[ 2] && Data[ 3] == s.Data[ 3] &&
			Data[ 4] == s.Data[ 4] && Data[ 5] == s.Data[ 5] && Data[ 6] == s.Data[ 6] && Data[ 7] == s.Data[ 7] &&
			Data[ 8] == s.Data[ 8] && Data[ 9] == s.Data[ 9] && Data[10] == s.Data[10] && Data[11] == s.Data[11] &&
			Data[12] == s.Data[12] && Data[13] == s.Data[13] && Data[14] == s.Data[14] && Data[15] == s.Data[15];
	}
	bool operator != (const MD5_SUM &s) const
	{
		return
			Data[ 0] != s.Data[ 0] || Data[ 1] != s.Data[ 1] || Data[ 2] != s.Data[ 2] || Data[ 3] != s.Data[ 3] ||
			Data[ 4] != s.Data[ 4] || Data[ 5] != s.Data[ 5] || Data[ 6] != s.Data[ 6] || Data[ 7] != s.Data[ 7] ||
			Data[ 8] != s.Data[ 8] || Data[ 9] != s.Data[ 9] || Data[10] != s.Data[10] || Data[11] != s.Data[11] ||
			Data[12] != s.Data[12] || Data[13] != s.Data[13] || Data[14] != s.Data[14] || Data[15] != s.Data[15];
	}
	bool operator < (const MD5_SUM &s) const;
	bool operator > (const MD5_SUM &s) const;
	bool operator <= (const MD5_SUM &s) const;
	bool operator >= (const MD5_SUM &s) const;
};

void MD5ToStr(tstring *Out, const MD5_SUM MD5);
bool StrToMD5(MD5_SUM *Out, const tstring &s);

/// Klasa obliczajca sum MD5 z kolejno podawanych blokw danych
/** Strumie tylko do zapisu.
Prawidowe uycie: MD5_Calc::Write(), MD5_Calc::Write() (lub inne funkcje zapisujce), ..., MD5_Calc::Finish().
Po wywoaniu Finish nie mona dalej zapisywa danych!
MD5_Calc::Reset() - rozpoczyna liczenie sumy od nowa. */
class MD5_Calc : public Stream
{
private:
	uint4 total[2];
	uint4 state[4];
	uint1 buffer[64];

	void Process(uint1 data[64]);

public:
	MD5_Calc();

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);

	/// Koczy obliczenia i zwraca policzon sum
	void Finish(MD5_SUM *Out);
	/// Rozpoczyna liczenie nowej sumy
	void Reset();

	// ======== Statyczne ========
	/// Po prostu oblicza sum kontroln z podanych danych
	static void Calc(MD5_SUM *Out, const void *Buf, uint4 BufLen);
};

/// Koduje lub dekoduje zapisywane/odczytywane bajty XOR podany bajt lub cig bajtw.
/** Mapuje bezporednio bajty na bajty strumienia do ktrego jest podczony,
nic nie buforuje, wic mona operowa te na strumieniu Stream. */
class XorCoder : public OverlayStream
{
private:
	std::vector<char> m_Buf;
	std::vector<char> m_Key;
	uint m_EncodeKeyIndex;
	uint m_DecodeKeyIndex;

public:
	XorCoder(Stream *a_Stream, char KeyByte);
	XorCoder(Stream *a_Stream, const void *KeyData, size_t KeyLength);
	XorCoder(Stream *a_Stream, const string &Key);

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Data, size_t Size);

	/// Resetuje liczniki tak eby kodowanie i dekodowanie odbywao si od pocztku klucza
	void Reset() { m_EncodeKeyIndex = m_DecodeKeyIndex = 0; }

	// ======== Statyczne ========
	/// Po prostu przetwarza podane dane
	static void Code(void *Out, const void *Data, size_t DataLength, const void *Key, size_t KeyLength);
	/// Po prostu przetwarza podane dane
	static void Code(string *Out, const string &Data, const string &Key);
#ifdef WIN32
	/// Po prostu przetwarza podane dane
	static void Code(wstring *Out, const wstring &Data, const wstring &Key);
#endif
};

/// Koduje strumie binarnych danych wejciowych na cig zoony ze znakw '0' i '1' (po 8 na kady bajt).
class BinEncoder : public OverlayStream
{
private:
	CharWriter m_CharWriter;

public:
	BinEncoder(Stream *a_Stream) : OverlayStream(a_Stream), m_CharWriter(a_Stream) { }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush() { m_CharWriter.Flush(); OverlayStream::Flush(); }

	// ======== Statyczne ========
	/// Po prostu koduje podane dane
	/*** \param[out] Out Bufor wyjciowy Out musi mie zaalokowane przynajmniej DataLength * 8 bajtw. */
	static void Encode(char *Out, const void *Data, size_t DataLength);
	/// Po prostu koduje podane dane
	/** \param[out] Out acuch wyjciowy Out nie musi mie adnego konkretnego rozmiaru -
	zostaje wyczyszczony i wypeniony od nowa. */
	static void Encode(string *Out, const void *Data, size_t DataLength);
};

/// Dekoduje cig znakw wejciowych zapisanych binarnie (znakami '0' i '1') na dane binarne.
class BinDecoder : public OverlayStream
{
private:
	CharReader m_CharReader;
	DECODE_TOLERANCE m_Tolerance;

public:
	BinDecoder(Stream *a_Stream, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);

	DECODE_TOLERANCE GetTolerance() { return m_Tolerance; }

	// ======== Implementacja Stream ========
	virtual size_t Read(void *Out, size_t Size);
	virtual bool End() { return m_CharReader.End(); }

	// ======== Statyczne ========
	/// Znajduje dugo zakodowanych danych i zwraca true.
	/** Jeli nie udao si ustali dugoci, zwraca false. */
	static bool DecodeLength(uint *OutLength, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static bool DecodeLength(uint *OutLength, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	/// Dekoduje dane binarne. Zwraca liczb zdekodowanych bajtw.
	/** Jeli nie udao si zdekodowa (jaki bd), zwraca MAXUINT4 (0xFFFFFFFF).
	\param[out] OutData musi by tablic zaalokowan przynajmniej na rozmiar ustalony wywoaniem DecodeLength()
	lub jeli nie chcesz wywoywa DecodeLength(), na dugo rwn dugoci danych wejciowych / 8. */
	static size_t Decode(void *OutData, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static size_t Decode(void *OutData, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
};

/// Koduje strumie binarnych danych wejciowych na zapis szesnastkowy (po 2 znaki na kady bajt).
class HexEncoder : public OverlayStream
{
private:
	CharWriter m_CharWriter;
	bool m_UpperCase;

public:
	HexEncoder(Stream *a_Stream, bool UpperCase = true) : OverlayStream(a_Stream), m_CharWriter(a_Stream), m_UpperCase(UpperCase) { }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush() { m_CharWriter.Flush(); OverlayStream::Flush(); }

	// ======== Statyczne ========
	/// Po prostu koduje podane dane
	/** \param[out] Out Bufor wyjciowy Out musi mie zaalokowane przynajmniej DataLength * 2 znakw. */
	static void Encode(char *Out, const void *Data, size_t DataLength, bool UpperCase = true);
	/// Po prostu koduje podane dane
	/** \param[out] Out acuch wyjciowy Out nie musi mie adnego konkretnego rozmiaru -
	zostaje wyczyszczony i wypeniony od nowa. */
	static void Encode(string *Out, const void *Data, size_t DataLength, bool UpperCase = true);
#ifdef WIN32
	static void Encode(wchar_t *Out, const void *Data, size_t DataLength, bool UpperCase = true);
	static void Encode(wstring *Out, const void *Data, size_t DataLength, bool UpperCase = true);
#endif
};

/// Dekoduje cig znakw wejciowych zapisanych szesnastkowo na dane binarne.
/** Akceptuje zarwno mae, jak i due litery. */
class HexDecoder : public OverlayStream
{
private:
	CharReader m_CharReader;
	DECODE_TOLERANCE m_Tolerance;

public:
	HexDecoder(Stream *a_Stream, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);

	DECODE_TOLERANCE GetTolerance() { return m_Tolerance; }

	// ======== Implementacja Stream ========
	virtual size_t Read(void *Out, size_t Size);
	virtual bool End() { return m_CharReader.End(); }

	// ======== Statyczne ========
	/// Znajduje dugo zakodowanych danych i zwraca true.
	/** Jeli nie udao si ustali dugoci, zwraca false. */
	static bool DecodeLength(uint *OutLength, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static bool DecodeLength(uint *OutLength, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	/// Dekoduje dane binarne. Zwraca liczb zdekodowanych bajtw.
	/** Jeli nie udao si zdekodowa (jaki bd), zwraca MAXUINT4 (0xFFFFFFFF).
	\param[out] OutData musi by tablic zaalokowan przynajmniej na rozmiar ustalony wywoaniem DecodeLength
	lub jeli nie chcesz wywoywa DecodeLength, na dugo rwn dugoci danych wejciowych / 2. */
	static size_t Decode(void *OutData, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	/// Dekoduje dane binarne. Zwraca liczb zdekodowanych bajtw.
	static size_t Decode(void *OutData, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
#ifdef WIN32
	static size_t Decode(void *OutData, const wstring &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static size_t Decode(void *OutData, const wchar_t *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
#endif
};

/// Koduje strumie binarnych danych wejciowych na cig znakw base64.
/**
- Zgodny z RFC 4648 "The Base16, Base32, and Base64 Data Encodings"
- Prawidowe dane wyjciowe powstaj dopiero po zakoczeniu zapisywania, ktre nastpuje w destruktorze.
  Jeli chcesz zakoczy wczeniej, wywoaj Base64Encoder::Finish(), ale potem nie mona ju nic dopisywa.
*/
class Base64Encoder : public OverlayStream
{
private:
	CharWriter m_CharWriter;
	bool m_Finished; // Czy dane zakoczone
	uint1 m_Buf[2];
	uint m_BufIndex;

	void DoFinish();

public:
	Base64Encoder(Stream *a_Stream) : OverlayStream(a_Stream), m_CharWriter(a_Stream), m_Finished(false), m_BufIndex(0) { }
	~Base64Encoder();

	void Finish() { if (!m_Finished) { DoFinish(); m_Finished = true; } }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush() { m_CharWriter.Flush(); OverlayStream::Flush(); }

	// ======== Statyczne ========
	/// Oblicza rozmiar bufora na zakodowane dane
	static size_t EncodeLength(size_t DataLength);
	/// Po prostu koduje podane dane
	/** \param[out] Out Bufor wyjciowy Out musi mie zaalokowane przynajmniej tyle bajtw ile wyliczyo EncodeLength.
	\return Zwraca liczb znakw zapisanych do Out. */
	static size_t Encode(char *Out, const void *Data, size_t DataLength);
	/// Po prostu koduje podane dane
	/** \param[out] Out acuch wyjciowy Out nie musi mie adnego konkretnego rozmiaru -
	zostaje wyczyszczony i wypeniony od nowa. */
	static size_t Encode(string *Out, const void *Data, size_t DataLength);
};

/// Dekoduje cig znakw wejciowych zapisanych w standardzie base64 na dane binarne.
/*
- Po napotkaniu kocwki zawierajcej znaki '=' funkcje nie parsuj ju dalej,
  ale to tylko optymalizacja i nie naley na tym polega.
- Sposb dziaania strumienia - na raz wczytywane s po 4 znaki i zamieniane na
  maksymalnie 3 bajty.
*/
class Base64Decoder : public OverlayStream
{
private:
	CharReader m_CharReader;
	DECODE_TOLERANCE m_Tolerance;
	// Bajty s w kolejnoci: [2][1][0]. Jeli jest mniej, to [1][0][-] albo [0][-][-].
	uint1 m_Buf[3];
	// Liczba zdekodowanych bajtw w buforze
	uint m_BufLength;
	// True, jeli sparsowano kocwk z '='
	bool m_Finished;

	// Czyta 4 znaki ze strumienia podczonego (0 albo 4, inaczej bd) i wypenia m_Buf maksymalnie 3 zdekodowanymi bajtami.
	// Ustawia m_BufLength i jeli trzeba, to m_Finished.
	bool ReadNextBuf();

	bool GetNextByte(uint1 *OutByte) { if (m_BufLength == 0) { if (!ReadNextBuf()) return false; } *OutByte = m_Buf[--m_BufLength]; return true; }

public:
	Base64Decoder(Stream *a_Stream, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);

	DECODE_TOLERANCE GetTolerance() { return m_Tolerance; }

	// ======== Implementacja Stream ========
	virtual size_t Read(void *Out, size_t Size);
	virtual bool End();

	// ======== Statyczne ========
	/// Znajduje dugo zakodowanych danych i zwraca true.
	/** Jeli nie udao si ustali dugoci, zwraca false. */
	static bool DecodeLength(uint *OutLength, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static bool DecodeLength(uint *OutLength, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	/// Dekoduje dane binarne. Zwraca liczb zdekodowanych bajtw.
	/** Jeli nie udao si zdekodowa (jaki bd), zwraca MAXUINT4 (0xFFFFFFFF).
	\param[out] OutData musi by tablic zaalokowan przynajmniej na rozmiar ustalony wywoaniem DecodeLength(). */
	static size_t Decode(void *OutData, const string &s, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
	static size_t Decode(void *OutData, const char *s, uint s_Length, DECODE_TOLERANCE Tolerance = DECODE_TOLERANCE_NONE);
};

/// Bufor koowy
/*
Dziaa szybko (na tyle na ile szybko moe dziaa strumie, ze
swoimi metodami wirtualnymi itd.), ale ma z gry ograniczon pojemno. W
przypadku przepenienia przy zapisie rzuca wyjtek.
*/
class RingBuffer : public Stream
{
public:
	RingBuffer(uint Capacity);
	virtual ~RingBuffer();

	/// Zwraca liczb bajtw w buforze
	uint GetSize() { return m_Size; }
	/// Zwraca pojemno bufora
	uint GetCapacity() { return m_Capacity; }
	/// Zwraca true, jeli bufor jest pusty
	bool IsEmpty() { return GetSize() == 0; }
	/// Zwraca true, jeli bufor jest peny
	bool IsFull() { return GetSize() == GetCapacity(); }

	// ======== Implementacja Stream ========
	virtual void Write(const void *Data, size_t Size);
	virtual size_t Read(void *Out, size_t MaxLength);
	virtual void MustRead(void *Out, size_t Length);
	virtual bool End();
	virtual size_t Skip(size_t MaxLength);

private:
	uint m_Capacity;
	uint m_Size;
	std::vector<char> m_Buf;
	uint m_BegIndex;
	uint m_EndIndex;
};


/// Kopiuje dane z jednego strumienia do drugiego - co najwyej MaxLength bajtw.
/** \return Zwraca liczb przekopiowanych danych. */
inline size_t Copy(Stream *Dst, Stream *Src, size_t MaxLength) { return Dst->CopyFrom(Src, MaxLength); }
/// Kopiuje dane z jednego strumienia do drugiego - dokadnie Length bajtw.
/** Jeli si nie uda przekopiowa tylu bajtw, rzuca bd. */
inline void MustCopy(Stream *Dst, Stream *Src, size_t Length) { return Dst->MustCopyFrom(Src, Length); }
/// Kopiuje dane z jednego strumienia do drugiego a do koca tego rdowego
inline void CopyToEnd(Stream *Dst, Stream *Src) { Dst->CopyFromToEnd(Src); }

//@}
// code_stream

} // namespace common

/** \addtogroup code_sthtostr */
//@{

template <>
struct SthToStr_obj<common::MD5_SUM>
{
	void operator () (tstring *Str, const common::MD5_SUM &Sth)
	{
		common::MD5ToStr(Str, Sth);
	}
	static inline bool IsSupported() { return true; }
};

template <>
struct StrToSth_obj<common::MD5_SUM>
{
	bool operator () (common::MD5_SUM *Sth, const tstring &Str)
	{
		return common::StrToMD5(Sth, Str);
	}
	static inline bool IsSupported() { return true; }
};

//@}
// code_sthtostr

#endif
