/** \file
\brief Biblioteka do wielowtkowoci i synchronizacji
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Threads \n
Elementy moduu: \ref code_threads
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_THREADS_H_
#define COMMON_THREADS_H_

namespace common
{

/** \addtogroup code_threads Modu Threads
Dokumentacja: \ref Module_Threads \n
Nagwek: Threads.hpp */
//@{

/// \internal
class Thread_pimpl;
/// \internal
class Mutex_pimpl;
/// \internal
class Semaphore_pimpl;
/// \internal
class Cond_pimpl;
/// \internal
class Barrier_pimpl;
/// \internal
class Event_pimpl;

/// Klasa bazowa wtku.
/**
- Odziedzicz po tej klasie, eby zdefiniowa wasny typ wtku. Nadpisa metod Thread::Run().
- Obiekt tej klasy reprezentuje pojedynczy wtek.
- Wtek mona odpali jeden raz - Thread::Start().
- Przed zniszczeniem obiektu na odpalony wtek trzeba poczeka - Thread::Join().
- Obiekt mona tworzy na stosie lub na stercie - obojtne.
- Obiekt, jeli alokowany new, trzeba zniszczy delete - sam si nie zniszczy.
*/
class Thread
{
	DECLARE_NO_COPY_CLASS(Thread)
	friend class Thread_pimpl;

private:
	scoped_ptr<Thread_pimpl> pimpl;
	
protected:
	/// Napisz swoj wersj z kodem, ktry ma si wykona w osobnym wtku.
	virtual void Run() = 0;

	/// Koczy dziaanie wtku
	/**
	- Wywoywa z funkcji Run().
	- Lepiej jednak normalnie zakoczy funkcj Run.
	  Stosowanie Exit() nie ma sensu i jest niebezpieczne dla zasobw.
	*/
	void Exit();
	/// Oddaje sterowanie innemu wtkowi
	/**
	- Wywoywa z funkcji Run().
	- Nazwa taka dziwna, bo jaki cham zdefiniowa makro o nazwie Yield!
	*/
	void Yield_();

public:
	Thread();
	virtual ~Thread();

	/// Uruchamiamia wtek.
	void Start();
	/// Czeka na zakoczenie wtku.
	/**
	- Jeli wtek ju si zakoczy, wychodzi od razu.
	- Jeli wtek jeszcze nie by odpalony albo ju raz byo testowane Join(), nic si nie dzieje.
	*/
	void Join();

	/// Brutalnie zabija wtek - nie uywa!
	void Kill();

// Rzeczy ktre nie wiem jak zrealizowa w pthreads - dziaaj tylko w Windows
#ifdef WIN32
	/// Zwraca true, jeli wtek jest w tej chwili uruchomiony i naprawd jeszcze dziaa
	bool IsRunning();
	/// Zatrzymuje wtek, wznawia wtek (posiada wewntrzny licznik zatrzyma)
	/**
	- Dziaa tylko midzy Start() a Join().
	- Raczej nie ma sensu tego uywa.
	*/
	void Pause();
	void Resume();

	/** \name Priorytet wtku
	Prioritytet wtku jest liczb cakowit. Te stae to niektre wartoci. Mog by te inne.
	Ustawianie i pobieranie priotytetu wtku dziaa tylko midzy Start() a Join(). */
	//@{
	static const int PRIORITY_IDLE;
	static const int PRIORITY_VERY_LOW;
	static const int PRIORITY_LOW;
	static const int PRIORITY_DEFAULT;
	static const int PRIORITY_HIGH;
	static const int PRIORITY_VERY_HIGH;
	static const int PRIORITY_REALTIME;
	void SetPriority(int Priority);
	int GetPriority();
	//@}
#endif
};

/// Muteks
/** Obiekt do wyznaczania sekcji krytycznej na wzajemnie wykluczajc si wyczno
jednego wtku.
*/
class Mutex
{
	DECLARE_NO_COPY_CLASS(Mutex)
	friend class Mutex_pimpl;
	friend class Cond;

private:
	scoped_ptr<Mutex_pimpl> pimpl;

public:
	/** \name Flagi bitowe do konstruktora */
	//@{
	/// Podaj, jeli bdziesz robi w ramach jednego wtku zagniedone blokowania.
	static const uint FLAG_RECURSIVE    = 0x01;
	/// Podaj, jeli bdziesz uywa metody TimeoutLock.
	static const uint FLAG_WAIT_TIMEOUT = 0x02;
	//@}

	Mutex(uint Flag);
	~Mutex();

	/// Blokuje muteks.
	void Lock();
	/// Odblokowuje muteks.
	void Unlock();
	/// Prbuje wej do sekcji krytycznej.
	/**
	- Jeli si uda, wchodzi blokujc muteks i zwraca true.
	- Jeli si nie uda bo muteks jest zablokowany przez inny wtek, nie blokuje sterowania tylko zawraca false.
	*/
	bool TryLock();
	/// Prbuje wej do sekcji krytycznej czekajc co najwyej podany czas.
	/**
	- Jeli si uda, wchodzi blokujc muteks i zwraca true.
	- Jeli si nie uda przez podany czas, bo muteks jest zablokowany przez inny wtek, zwraca false.
	*/
	bool TimeoutLock(uint Milliseconds);
};

/// Klasa pomagajca blokowa muteks
/**
- W konstruktorze blokuje, w destruktorze odblokowuje.
- Uywaj jej zamiast metod Mutex::Lock i Mutex::Unlock kiedy tylko mona.
*/
class MutexLock
{
private:
	Mutex *m_Mutex;
	DECLARE_NO_COPY_CLASS(MutexLock)

public:
	MutexLock(Mutex *m) : m_Mutex(m) { m->Lock(); }
	~MutexLock() { m_Mutex->Unlock(); }
};

/**
Dla jeszcze wiszej wygody, zamiast tworzy obiekt klasy MutexLock wystarczy na
pocztku funkcji czy dowolnego bloku { } postawi to makro (M ma by typu
Mutex*).
*/
#define MUTEX_LOCK(M) MutexLock __mutex_lock_obj(M);

/// Semafor zliczajcy
/**
- Posiada wewntrznie nieujemn warto cakowit, ktr mona sobie wyobrazi
  jako liczb dostpnych zasobw. Nie ma do niej dostpu bezporednio.
- Warto maksymalna tej liczby nie jest okrelona, jest jaka dua.
*/
class Semaphore
{
	DECLARE_NO_COPY_CLASS(Semaphore)

private:
	scoped_ptr<Semaphore_pimpl> pimpl;

public:
	Semaphore(uint InitialValue);
	~Semaphore();

	/// Opuszcza semafor - zmniejsza jego warto o jeden.
	/**
	Jeli semafor jest zerowy, czeka a kto go podniesie zanim go opuci i zwrci sterowanie.
	*/
	void P();
	/// Prbuje opuci semafor - zmniejszy jego warto o jeden.
	/**
	- Jeli si uda, zmniejsza i zwraca true.
	- Jeli si nie uda bo warto wynosi 0, nie blokuje sterowania tylko zawraca false.
	*/
	bool TryP();
	/// Prbuje opuci semafor czekajc co najwyej podany czas.
	/**
	- Jeli si uda, zmniejsza i zwraca true.
	- Jeli si nie uda przez podany czas, bo warto wynosi 0, zwraca false.
	*/
	bool TimeoutP(uint Milliseconds);

	/// Podnosi semafor zwikszajc go o 1. Nigdy nie blokuje.
	void V();
	/// Podnosi semafor zwikszajc go o ReleaseCount. Nigdy nie blokuje.
	void V(uint ReleaseCount);
};

/// Zmienna warunkowa
/**
Nie pamita adnego swojego stanu. Pozwala za to zawiesi wtek eby na niej
poczeka (Wait()) oraz wznowi niezdefiniowany jeden (Signal()) lub wszystkie
(Broadcast()) czekajce na niej wtki. Sensu nabiera dopiero kiedy zrozumie si
sposb jej uycia w kontekcie wzorca monitora do sprawdzania warunku.
*/
class Cond
{
	DECLARE_NO_COPY_CLASS(Cond)

private:
	scoped_ptr<Cond_pimpl> pimpl;

public:
	Cond();
	~Cond();

	/// Odblokowuje muteks, zawiesza wtek czekajc na sygna, po doczekaniu si blokuje muteks z powrotem.
	/** Muteks m musi by zablokowany w chwili tego wywoania. */
	void Wait(Mutex *m);
	/// Odblokowuje muteks, zawiesza wtek czekajc na sygna przez co najwyej podany czas, po doczekaniu si blokuje muteks z powrotem.
	/**
	- Jeli w tym czasie nastpio przerwanie czekania przez sygna, zwraca true.
	- Muteks m musi by zablokowany w chwili tego wywoania.
	*/
	bool TimeoutWait(Mutex *m, uint Milliseconds);

	/// Odwiesza przypadkowy jeden czekajcy wtek.
	/**
	- Jeli aden wtek nie czeka, nie robi nic.
	- Nie jest wymagane eby jaki muteks musia by zablokowany w chwili tego wywoania.
	*/
	void Signal();
	/// Odwiesza wszystkie czekajce wtki.
	/**
	- Jeli aden wtek nie czeka, nie robi nic.
	- Nie jest wymagane eby jaki muteks musia by zablokowany w chwili tego wywoania.
	*/
	void Broadcast();
};

/// Bariera
/**
Wtki ktre wywouj Wait() blokuj si czekajc, a si ich uzbiera tyle
czekajcych, ile wynosi przekazany do konstruktora NumThreads. Kiedy ostatni z
nich wywoa Wait(), wszystkie ruszaj dalej.
*/
class Barrier
{
	DECLARE_NO_COPY_CLASS(Barrier)

private:
	scoped_ptr<Barrier_pimpl> pimpl;

public:
	Barrier(uint NumThreads);
	~Barrier();

	void Wait();
};

/// Zdarzenie
/**
Ten obiekt, wzorowany na WinAPI. Jest troch podobny do zmiennej warunkowej,
ale ma swj stan - true lub false.
*/
class Event
{
	DECLARE_NO_COPY_CLASS(Event);

private:
	scoped_ptr<Event_pimpl> pimpl;

public:
	enum TYPE
	{
		/** Zdarzenie samo przywrci si do false, kiedy jaki jeden wtek doczeka si true w wywoaniu Wait.
		SetEvent() wznawia jeden wtek czekajacy na Wait. */
		TYPE_AUTO_RESET,
		/** Zdarzenie samo nie przywrci si do false.
		SetEvent() wznawia wszystkie wtki czekajce na Wait. */
		TYPE_MANUAL_RESET,
	};

	Event(bool InitialState, TYPE Type);
	~Event();

	/// Zmienia stan na true.
	/**
	- Event::TYPE_AUTO_RESET: Jeli jakie wtki czekay, wznawia jeden i przestawia z powrotem na false.
	- Event::TYPE_MANUAL_RESET: Jeli jakie wtki czekay, wznawia je wszystkie.
	*/
	void Set();
	/// Zmienia stan na false.
	void Reset();

	/// Czeka, a stan bdzie true.
	/**
	- Jeli ju jest, zwraca sterowanie od razu.
	- Event::TYPE_AUTO_RESET: Kiedy si doczeka, przestawia go z powrotem na false.
	*/
	void Wait();
	/// Zwraca stan
	/** Event::TYPE_AUTO_RESET: Jeli by true, przestawia go z powrotem na false. */
	bool Test();
	/// Czeka, a stan bdzie true przez co najwyej podany czas.
	/**
	- Jeli jest albo si doczeka a bdzie, zwraca true.
	- Jeli si nie doczeka przez podany czas (stan jest cay czas false), zwraca false.
	- Event::TYPE_AUTO_RESET: Kiedy si doczeka na stan true, przestawia go z powrotem na false.
	*/
	bool TimeoutWait(uint Milliseconds);
};

//@}
// code_threads

} // namespace common

#endif
