/** \file
\brief Parser skadni jzyka podobnego do C/C++
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Tokenizer \n
Elementy moduu: \ref code_tokenizer
*/
#include "Base.hpp"
#include <algorithm>
#include <map>
#include "Stream.hpp"
#include "Tokenizer.hpp"


namespace common
{

TokenizerError::TokenizerError(size_t Char, size_t Row, size_t Col, const tstring &Msg, const tstring &File, int a_Line) :
	m_Char(Char),
	m_Row(Row),
	m_Col(Col)
{
	if (Msg.empty())
		Push(Format(_T("Tokenizer(row=#, col=#, char=#)")) % m_Row % m_Col % m_Char, __TFILE__, __LINE__);
	else
		Push(Format(_T("Tokenizer(row=#, col=#, char=#) #")) % m_Row % m_Col % m_Char % Msg, __TFILE__, __LINE__);
}


//HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH
// Klasa Tokenizer_pimpl

typedef std::map<tstring, uint> KEYWORDS_MAP;

class Tokenizer_pimpl
{
public:
	// Ktre znaki s symbolami
	bool m_Symbols[256];

	// Flagi
	bool m_FlagTokenEOL;
	bool m_FlagMultilineStrings;

	// Strumie do pobierania (jeden z tych trzech strumieni jest != NULL)
	// - Z konstruktora char*, wchar_t*, wstring - strumie pamiciowy, mj wasny
	scoped_ptr<MemoryStream> m_MemoryStream;
	// - Z konstruktora string - strumie acuchowy, mj wasny
	scoped_ptr<StringStream> m_StringStream;
	// - Z konstruktora Stream* - strumie dowolny, zewntrzny
	Stream *m_ExternalStream;
	// CharReader do jednego z powyszych strumieni
	scoped_ptr<TCharReader> m_CharReader;

	// WARSTWA I - wczytywanie znakw
	size_t m_CurrChar, m_CurrRow, m_CurrCol;
	bool m_L1End; // albo to jest true...
	tchar m_L1Char; // ...albo tu jest ostatnio wczytany znak
	void L1Next();

	// Zarejestrowane dane
	KEYWORDS_MAP m_Keywords;

	// Informacje na temat ostatnio odczytanego tokena
	size_t m_LastChar, m_LastRow, m_LastCol;
	Tokenizer::TOKEN m_LastToken;
	tstring m_LastString; // wypeniany zawsze przy odczytaniu nastpnego tokena
	uint m_LastId; // wypeniany przy TOKEN_KEYWORD

	~Tokenizer_pimpl() { }

	void InitSymbols();

	tchar ParseStringChar();
	void Parse();

	template <typename T> bool GetUint(T *Out);
	template <typename T> bool GetInt(T *Out);
	template <typename T> T MustGetUint();
	template <typename T> T MustGetInt();
};

void Tokenizer_pimpl::L1Next()
{
	m_L1End = !m_CharReader->ReadChar(&m_L1Char);

	if (!m_L1End)
	{
		// Liczniki
		m_CurrChar++;
		if (m_L1Char == _T('\n'))
		{
			m_CurrRow++;
			m_CurrCol = 0;
		}
		else
			m_CurrCol++;

		// Sekwencja "\\\n" lub "\\\r\n" jest pomijana - amanie linii za pomoc znaku \ ...
		if (m_L1Char == _T('\\'))
		{
			tchar Ch;
			if (m_CharReader->PeekChar(&Ch) && Ch == _T('\r'))
			{
				m_CharReader->SkipChar();
				m_CurrChar++;
			}
			if (m_CharReader->PeekChar(&Ch) && Ch == _T('\n'))
			{
				m_CharReader->SkipChar();
				m_CurrChar++;
				m_CurrRow++;
				m_CurrCol = 0;
				L1Next();
			}
		}
	}
}

void Tokenizer_pimpl::InitSymbols()
{
	std::fill(&m_Symbols[0], &m_Symbols[256], false);

	const tchar * DEFAULT_SYMBOLS = _T("`~!@#$%^&*()=[]{};:,.<>?\\|");

#ifdef WIN32
	uint SymbolCount = _tcslen(DEFAULT_SYMBOLS);
#else
	uint SymbolCount = strlen(DEFAULT_SYMBOLS);
#endif
	for (uint i = 0; i < SymbolCount; i++)
		m_Symbols[(uint1)DEFAULT_SYMBOLS[i]] = true;
}

tchar Tokenizer_pimpl::ParseStringChar()
{
	// Musi by tu ju sprawdzone, e nie jest koniec oraz e m_L1Char nie jest znakiem koczcym acuch/znak
	assert(!m_L1End);

	// Escape sequence
	if (m_L1Char == _T('\\'))
	{
		L1Next();
		if (m_L1End)
			throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected and of data inside escape sequence."), __TFILE__, __LINE__);
		// Staa szesnastkowa
		if (m_L1Char == 'x')
		{
			uint1 Number1, Number2;
			L1Next();

			// Nastpne dwa znaki

			if (m_L1End)
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected and of data inside escape sequence."), __TFILE__, __LINE__);
			Number1 = HexDigitToNumber(m_L1Char);
			L1Next();

			if (m_L1End)
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected and of data inside escape sequence."), __TFILE__, __LINE__);
			Number2 = HexDigitToNumber(m_L1Char);
			L1Next();

			if (Number1 == 0xFF || Number2 == 0xFF)
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Invalid hexadecimal escape sequence."), __TFILE__, __LINE__);

			return (tchar)(Number1 << 4 | Number2);
		}
		// Zwyka staa
		else
		{
			tchar R;
			switch (m_L1Char)
			{
			case _T('n'): R = _T('\n'); break;
			case _T('t'): R = _T('\t'); break;
			case _T('v'): R = _T('\v'); break;
			case _T('b'): R = _T('\b'); break;
			case _T('r'): R = _T('\r'); break;
			case _T('f'): R = _T('\f'); break;
			case _T('a'): R = _T('\a'); break;
			case _T('0'): R = _T('\0'); break;
			case _T('\\'):
			case _T('?'):
			case _T('"'):
			case _T('\''):
				R = m_L1Char;
				// Zostaje to co jest
				break;
			default:
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unknown escape sequence: \\") + CharToStrR(m_L1Char), __TFILE__, __LINE__);
			}
			L1Next();
			return R;
		}
	}
	else
	{
		tchar R = m_L1Char;
		L1Next();
		return R;
	}
}

void Tokenizer_pimpl::Parse()
{
	for (;;)
	{
		m_LastChar = m_CurrChar;
		m_LastRow = m_CurrRow;
		m_LastCol = m_CurrCol;

		// EOF
		if (m_L1End)
		{
			m_LastString.clear();
			m_LastToken = Tokenizer::TOKEN_EOF;
			return;
		}

		// Symbol
		if (m_Symbols[(uint1)m_L1Char])
		{
			CharToStr(&m_LastString, m_L1Char);
			m_LastToken = Tokenizer::TOKEN_SYMBOL;
			L1Next();
			return;
		}
		// Biay znak
		else if (m_L1Char == _T(' ') || m_L1Char == _T('\t') || m_L1Char == _T('\r') || m_L1Char == _T('\v'))
		{
			L1Next();
			continue;
		}
		// Nagwek Unicode
#ifdef _UNICODE
		else if (m_L1Char == L'\xFFFE')
		{
			L1Next();
			continue;
		}
#endif
		// Koniec wiersza
		else if (m_L1Char == _T('\n'))
		{
			// Jako token
			if (m_FlagTokenEOL)
			{
				m_LastString = _T("\n");
				m_LastToken = Tokenizer::TOKEN_EOL;
				L1Next();
				return;
			}
			// Jako biay znak
			else
			{
				L1Next();
				continue;
			}
		}
		// Staa znakowa
		else if (m_L1Char == _T('\''))
		{
			// Nastpny znak
			L1Next();
			if (m_L1End)
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected end of data inside character constant."), __TFILE__, __LINE__);
			// Znak (sekwencja ucieczki lub zwyky)
			CharToStr(&m_LastString, ParseStringChar());
			// Nastpny znak - to musi by zakoczenie '
			if (m_L1End)
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected end of data inside character constant."), __TFILE__, __LINE__);
			if (m_L1Char != _T('\''))
				throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("End of character constant expected: '"), __TFILE__, __LINE__);
			m_LastToken = Tokenizer::TOKEN_CHAR;
			L1Next();
			return;
		}
		// Staa acuchowa
		else if (m_L1Char == _T('"'))
		{
			L1Next();
			m_LastString.clear();
			for (;;)
			{
				// Nastpny znak
				if (m_L1End)
					throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected end of data inside string constant."), __TFILE__, __LINE__);
				// Koniec acucha
				if (m_L1Char == _T('"'))
				{
					L1Next();
					break;
				}
				// Koniec wiersza, podczas gdy nie moe go by
				else if ( (m_L1Char == _T('\r') || m_L1Char == _T('\n')) && !m_FlagMultilineStrings )
					throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unexpected end of data inside string constant."), __TFILE__, __LINE__);
				// Znak (sekwencja ucieczki lub zwyky)
				else
					m_LastString += ParseStringChar();
			}
			m_LastToken = Tokenizer::TOKEN_STRING;
			return;
		}
		// Identyfikator lub sowo kluczowe
		else if ( (m_L1Char <= _T('Z') && m_L1Char >= _T('A')) || (m_L1Char <= _T('z') && m_L1Char >= _T('a')) || m_L1Char == _T('_') )
		{
			CharToStr(&m_LastString, m_L1Char);
			L1Next();
			// Wczytaj nastpne znaki
			while ( !m_L1End &&
				( (m_L1Char <= _T('Z') && m_L1Char >= _T('A')) || (m_L1Char <= _T('z') && m_L1Char >= _T('a')) || (m_L1Char <= _T('9') && m_L1Char >= _T('0')) || m_L1Char == _T('_') ) )
			{
				m_LastString += m_L1Char;
				L1Next();
			}
			// Znajd sowo kluczowe
			KEYWORDS_MAP::iterator it = m_Keywords.find(m_LastString);
			if (it == m_Keywords.end())
				m_LastToken = Tokenizer::TOKEN_IDENTIFIER;
			else
			{
				m_LastToken = Tokenizer::TOKEN_KEYWORD;
				m_LastId = it->second;
			}
			return;
		}
		// Znak '/' moe rozpoczyna komentarz // , /* lub by zwykym symbolem
		else if (m_L1Char == _T('/'))
		{
			L1Next();
			if (!m_L1End)
			{
				// Pocztek komentarza //
				if (m_L1Char == _T('/'))
				{
					L1Next();
					// Pomi znaki do koca wiersza lub koca dokumentu
					while (!m_L1End && m_L1Char != _T('\n'))
						L1Next();
					// Dalsze poszukiwanie tokena
					continue;
				}
				// Pocztek komentarza /*
				else if (m_L1Char == _T('*'))
				{
					L1Next();
					bool WasAsterisk = false;
					// Czytaj kolejne znaki
					for (;;)
					{
						if (m_L1End)
							throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Unfinished comment /* */"), __TFILE__, __LINE__);
						
						if (m_L1Char == _T('*'))
							WasAsterisk = true;
						else if (m_L1Char == _T('/') && WasAsterisk)
							// Koniec komentarza
							break;
						else
							WasAsterisk = false;

						L1Next();
					}
					// Dalsze poszukiwanie tokena
					L1Next();
					continue;
				}
			}
			// Zwyky symbol '/'
			CharToStr(&m_LastString, _T('/'));
			m_LastToken = Tokenizer::TOKEN_SYMBOL;
			L1Next();
			return;
		}
		// Liczba cakowita lub zmiennoprzecinkowa
		else if ( (m_L1Char <= _T('9') && m_L1Char >= _T('0')) || m_L1Char == _T('+') || m_L1Char == _T('-') )
		{
			// Pena poprawna skadnia liczb to mniej wicej (wyraenie regularne, \o \d \h to cyfra w danym systemie):
			// [+-]?(0[xX]\h+|0\o+|(\d+\.\d*|\d*\.\d+|\d+)([dDeE][+-]?\d+)?)
			// Ale e nie chce mi si tego a tak dokadnie parsowa, to zaakceptuj tutaj dowolny cig
			// dowolnych cyfr, liter i znakw '.', '+' i '-' - ewentualny bd skadni i tak wyjdzie, tylko e pniej.
			// Liczba jest na pewno cakowita, jeli zaczyna si od "0x" lub "0X" (nowo! tu by bd i 0xEE rozpoznawao jako zmiennoprzecinkowa!)
			// Liczba jest zmiennoprzecinkowa wtedy, kiedy zawiera jeden ze znakw: [.dDeE]
			CharToStr(&m_LastString, m_L1Char);
			L1Next();
			bool IsFloat = false;
			while (!m_L1End && (
				(m_L1Char <= _T('9') && m_L1Char >= _T('0')) ||
				(m_L1Char <= _T('Z') && m_L1Char >= _T('A')) ||
				(m_L1Char <= _T('z') && m_L1Char >= _T('a')) ||
				m_L1Char == _T('+') || m_L1Char == _T('-') || m_L1Char == _T('.') ) )
			{
				if (m_L1Char == _T('.') || m_L1Char == _T('d') || m_L1Char == _T('D') || m_L1Char == _T('e') || m_L1Char == _T('E'))
					IsFloat = true;
				m_LastString += m_L1Char;
				L1Next();
			}
			if (m_LastString.length() > 1 && m_LastString[0] == _T('0') && (m_LastString[1] == _T('x') || m_LastString[1] == _T('X')))
				m_LastToken = Tokenizer::TOKEN_INTEGER;
			else
				m_LastToken = (IsFloat ? Tokenizer::TOKEN_FLOAT : Tokenizer::TOKEN_INTEGER);
			return;
		}
		// Nieznany znak
		else
			throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, Format(_T("Unrecognized character (code=0x#): \'#\'.")) % IntToStrR(TCHAR_TO_INT(m_L1Char), 16) % m_L1Char, __TFILE__, __LINE__);
	}
}

template <typename T>
bool Tokenizer_pimpl::GetUint(T *Out)
{
	if (m_LastString == _T("0"))
	{
		*Out = T();
		return true;
	}

	// semkowo lub szesnastkowo
	if (!m_LastString.empty() && m_LastString[0] == _T('0'))
	{
		// Szesnastkowo
		if (m_LastString.length() > 1 && (m_LastString[1] == _T('x') || m_LastString[1] == _T('X')))
			return ( StrToUint<T>(Out, m_LastString.substr(2), 16) == 0 );
		// semkowo
		else
			return ( StrToUint<T>(Out, m_LastString.substr(1), 8) == 0 );
	}
	// Dziesitnie
	else
		return ( StrToUint<T>(Out, m_LastString) == 0 );
}

template <typename T>
bool Tokenizer_pimpl::GetInt(T *Out)
{
	if (m_LastString == _T("0") || m_LastString == _T("-0") || m_LastString == _T("+0"))
	{
		*Out = T();
		return true;
	}

	// Pocztkowy '+' lub '-'
	if (!m_LastString.empty() && (m_LastString[0] == _T('+') || m_LastString[0] == _T('-')))
	{
		// Plus - tak jakby go nie byo, ale wszystko jest o znak dalej
		if (m_LastString[0] == _T('+'))
		{
			// semkowo lub szesnastkowo
			if (m_LastString.length() > 1 && m_LastString[1] == _T('0'))
			{
				// Szesnastkowo
				if (m_LastString.length() > 2 && (m_LastString[2] == _T('x') || m_LastString[2] == _T('X')))
					return ( StrToInt<T>(Out, m_LastString.substr(3), 16) == 0 );
				// semkowo
				else
					return ( StrToInt<T>(Out, m_LastString.substr(2), 8) == 0 );
			}
			// Dziesitnie
			else
				return ( StrToInt<T>(Out, m_LastString.substr(1)) == 0 );
		}
		// Minus - uwzgldnij go
		else
		{
			// semkowo lub szesnastkowo
			if (m_LastString.length() > 1 && m_LastString[1] == _T('0'))
			{
				// Szesnastkowo
				if (m_LastString.length() > 2 && (m_LastString[2] == _T('x') || m_LastString[2] == _T('X')))
					return ( StrToInt<T>(Out, _T("-") + m_LastString.substr(3), 16) == 0 );
				// semkowo
				else
					return ( StrToInt<T>(Out, _T("-") + m_LastString.substr(2), 8) == 0 );
			}
			// Dziesitnie
			else
				return ( StrToInt<T>(Out, _T("-") + m_LastString.substr(1)) == 0 );
		}
	}
	else
	{
		// semkowo lub szesnastkowo
		if (!m_LastString.empty() && m_LastString[0] == _T('0'))
		{
			// Szesnastkowo
			if (m_LastString.length() > 1 && (m_LastString[1] == _T('x') || m_LastString[1] == _T('X')))
				return ( StrToInt<T>(Out, m_LastString.substr(2), 16) == 0 );
			// semkowo
			else
				return ( StrToInt<T>(Out, m_LastString.substr(1), 8) == 0 );
		}
		// Dziesitnie
		else
			return ( StrToInt<T>(Out, m_LastString) == 0 );
	}
}

template <typename T>
T Tokenizer_pimpl::MustGetUint()
{
	T R;
	if (!GetUint(&R))
		throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Cannot convert tstring to unsigned integer number."), __TFILE__, __LINE__);
	return R;
}

template <typename T>
T Tokenizer_pimpl::MustGetInt()
{
	T R;
	if (!GetInt(&R))
		throw TokenizerError(m_LastChar, m_LastRow, m_LastCol, _T("Cannot convert tstring to signed integer number."), __TFILE__, __LINE__);
	return R;
}


//HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH
// Klasa Tokenizer

Tokenizer::Tokenizer(const tchar *Input, size_t NumChars, uint Flags) :
	pimpl(new Tokenizer_pimpl())
{
	pimpl->InitSymbols();

	pimpl->m_FlagTokenEOL = ((Flags & FLAG_TOKEN_EOL) != 0);
	pimpl->m_FlagMultilineStrings = ((Flags & FLAG_MULTILINE_STRINGS) != 0);

	pimpl->m_MemoryStream.reset(new MemoryStream(NumChars * sizeof(tchar), const_cast<tchar*>(Input)));
	pimpl->m_StringStream.reset();
	pimpl->m_ExternalStream = NULL;
	pimpl->m_CharReader.reset(new TCharReader(pimpl->m_MemoryStream.get()));

	pimpl->m_CurrChar = 0; pimpl->m_CurrRow = 1; pimpl->m_CurrCol = 0;
	pimpl->L1Next();

	pimpl->m_LastChar = 0; pimpl->m_LastRow = 1; pimpl->m_LastCol = 0;
	pimpl->m_LastToken = TOKEN_EOF;
	pimpl->m_LastId = 0;
}

Tokenizer::Tokenizer(const tstring *Input, uint Flags) :
	pimpl(new Tokenizer_pimpl())
{
	pimpl->InitSymbols();

	pimpl->m_FlagTokenEOL = ((Flags & FLAG_TOKEN_EOL) != 0);
	pimpl->m_FlagMultilineStrings = ((Flags & FLAG_MULTILINE_STRINGS) != 0);

	pimpl->m_ExternalStream = NULL;

#ifdef _UNICODE
	pimpl->m_MemoryStream.reset(new MemoryStream(Input->length() * sizeof(tchar), const_cast<tchar*>(Input->data())));
	pimpl->m_StringStream.reset();
	pimpl->m_CharReader.reset(new TCharReader(pimpl->m_MemoryStream.get()));
#else
	pimpl->m_MemoryStream.reset();
	pimpl->m_StringStream.reset(new StringStream(const_cast<tstring*>(Input)));
	pimpl->m_CharReader.reset(new TCharReader(pimpl->m_StringStream.get()));
#endif

	pimpl->m_CurrChar = 0; pimpl->m_CurrRow = 1; pimpl->m_CurrCol = 0;
	pimpl->L1Next();

	pimpl->m_LastChar = 0; pimpl->m_LastRow = 1; pimpl->m_LastCol = 0;
	pimpl->m_LastToken = TOKEN_EOF;
	pimpl->m_LastId = 0;
}

Tokenizer::Tokenizer(Stream *Input, uint Flags) :
	pimpl(new Tokenizer_pimpl())
{
	pimpl->InitSymbols();

	pimpl->m_FlagTokenEOL = ((Flags & FLAG_TOKEN_EOL) != 0);
	pimpl->m_FlagMultilineStrings = ((Flags & FLAG_MULTILINE_STRINGS) != 0);

	pimpl->m_MemoryStream.reset();
	pimpl->m_StringStream.reset();
	pimpl->m_ExternalStream = Input;
	pimpl->m_CharReader.reset(new TCharReader(Input));

	pimpl->m_CurrChar = 0; pimpl->m_CurrRow = 1; pimpl->m_CurrCol = 0;
	pimpl->L1Next();

	pimpl->m_LastChar = 0; pimpl->m_LastRow = 1; pimpl->m_LastCol = 0;
	pimpl->m_LastToken = TOKEN_EOF;
	pimpl->m_LastId = 0;
}

Tokenizer::~Tokenizer()
{
	pimpl->m_CharReader.reset();
	pimpl->m_ExternalStream = NULL;
	pimpl->m_StringStream.reset();
	pimpl->m_MemoryStream.reset();

	pimpl.reset();
}

void Tokenizer::RegisterKeyword(uint Id, const tstring &Keyword)
{
	pimpl->m_Keywords.insert(KEYWORDS_MAP::value_type(Keyword, Id));
}

void Tokenizer::RegisterKeywords(const tchar **Keywords, size_t KeywordCount)
{
	for (uint i = 0; i < KeywordCount; i++)
		RegisterKeyword(i, Keywords[i]);
}

void Tokenizer::Next()
{
	pimpl->Parse();
}

Tokenizer::TOKEN Tokenizer::GetToken()
{
	return pimpl->m_LastToken;
}

size_t Tokenizer::GetCharNum()
{
	return pimpl->m_LastChar;
}

size_t Tokenizer::GetRowNum()
{
	return pimpl->m_LastRow;
}

size_t Tokenizer::GetColNum()
{
	return pimpl->m_LastCol;
}

const tstring & Tokenizer::GetString()
{
	return pimpl->m_LastString;
}

tchar Tokenizer::GetChar()
{
	assert(GetToken() == TOKEN_CHAR || GetToken() == TOKEN_SYMBOL);

	return pimpl->m_LastString[0];
}

uint Tokenizer::GetId()
{
	assert(GetToken() == TOKEN_KEYWORD);

	return pimpl->m_LastId;
}

bool Tokenizer::GetUint1(uint1 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetUint<uint1>(Out); }
bool Tokenizer::GetUint2(uint2 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetUint<uint2>(Out); }
bool Tokenizer::GetUint4(uint4 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetUint<uint4>(Out); }
bool Tokenizer::GetUint8(uint8 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetUint<uint8>(Out); }

bool Tokenizer::GetInt1(int1 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetInt<int1>(Out); }
bool Tokenizer::GetInt2(int2 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetInt<int2>(Out); }
bool Tokenizer::GetInt4(int4 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetInt<int4>(Out); }
bool Tokenizer::GetInt8(int8 *Out) { AssertToken(TOKEN_INTEGER); return pimpl->GetInt<int8>(Out); }

uint1 Tokenizer::MustGetUint1() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetUint<uint1>(); }
uint2 Tokenizer::MustGetUint2() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetUint<uint2>(); }
uint4 Tokenizer::MustGetUint4() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetUint<uint4>(); }
uint8 Tokenizer::MustGetUint8() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetUint<uint8>(); }

int1 Tokenizer::MustGetInt1() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetInt<int1>(); }
int2 Tokenizer::MustGetInt2() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetInt<int2>(); }
int4 Tokenizer::MustGetInt4() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetInt<int4>(); }
int8 Tokenizer::MustGetInt8() { AssertToken(TOKEN_INTEGER); return pimpl->MustGetInt<int8>(); }

bool Tokenizer::GetFloat(float *Out)
{
	AssertToken(TOKEN_INTEGER, TOKEN_FLOAT);

	return ( StrToFloat(Out, pimpl->m_LastString) == 0 );
}

bool Tokenizer::GetDouble(double *Out)
{
	AssertToken(TOKEN_INTEGER, TOKEN_FLOAT);

	return ( StrToDouble(Out, pimpl->m_LastString) == 0 );
}

float Tokenizer::MustGetFloat()
{
	float R;
	if (!GetFloat(&R))
		throw TokenizerError(pimpl->m_LastChar, pimpl->m_LastRow, pimpl->m_LastCol, _T("Cannot convert tstring to float number."), __TFILE__, __LINE__);
	return R;
}

double Tokenizer::MustGetDouble()
{
	double R;
	if (!GetDouble(&R))
		throw TokenizerError(pimpl->m_LastChar, pimpl->m_LastRow, pimpl->m_LastCol, _T("Cannot convert tstring to double number."), __TFILE__, __LINE__);
	return R;
}

void Tokenizer::CreateError()
{
	throw TokenizerError(pimpl->m_LastChar, pimpl->m_LastRow, pimpl->m_LastCol, _T("Unknown error"), __TFILE__, __LINE__);
}

void Tokenizer::CreateError(const tstring &Msg)
{
	throw TokenizerError(pimpl->m_LastChar, pimpl->m_LastRow, pimpl->m_LastCol, Msg, __TFILE__, __LINE__);
}

bool Tokenizer::QueryToken(Tokenizer::TOKEN Token)
{
	return (GetToken() == Token);
}

bool Tokenizer::QueryToken(TOKEN Token1, TOKEN Token2)
{
	return (GetToken() == Token1 || GetToken() == Token2);
}

bool Tokenizer::QueryEOF()
{
	return QueryToken(TOKEN_EOF);
}

bool Tokenizer::QueryEOL()
{
	return QueryToken(TOKEN_EOL);
}

bool Tokenizer::QuerySymbol(tchar Symbol)
{
	return (GetToken() == TOKEN_SYMBOL && GetChar() == Symbol);
}

bool Tokenizer::QueryIdentifier(const tstring &Identifier)
{
	return (GetToken() == TOKEN_IDENTIFIER && GetString() == Identifier);
}

bool Tokenizer::QueryKeyword(uint KeywordId)
{
	return (GetToken() == TOKEN_KEYWORD && GetId() == KeywordId);
}

bool Tokenizer::QueryKeyword(const tstring &Keyword)
{
	return (GetToken() == TOKEN_KEYWORD && GetString() == Keyword);
}

void Tokenizer::AssertToken(Tokenizer::TOKEN Token)
{
	if (Token != GetToken())
		CreateError(_T("Expected: ") + tstring(GetTokenName(Token)));
}

void Tokenizer::AssertToken(TOKEN Token1, TOKEN Token2)
{
	if (Token1 != GetToken() && Token2 != GetToken())
		CreateError(_T("Expected: ") + tstring(GetTokenName(Token1)) + _T(" or ") + GetTokenName(Token2));
}

void Tokenizer::AssertEOF()
{
	AssertToken(TOKEN_EOF);
}

void Tokenizer::AssertEOL()
{
	AssertToken(TOKEN_EOL);
}

void Tokenizer::AssertSymbol(tchar Symbol)
{
	if (GetToken() != TOKEN_SYMBOL || GetChar() != Symbol)
		CreateError(_T("Expected symbol: ") + CharToStrR(Symbol));
}

void Tokenizer::AssertIdentifier(const tstring &Identifier)
{
	if (GetToken() != TOKEN_IDENTIFIER || GetString() != Identifier)
		CreateError(_T("Expected identifier: ") + Identifier);
}

void Tokenizer::AssertKeyword(uint KeywordId)
{
	if (GetToken() != TOKEN_KEYWORD || GetId() != KeywordId)
		CreateError(_T("Expected keyword number ") + UintToStrR(KeywordId));
}

void Tokenizer::AssertKeyword(const tstring &Keyword)
{
	if (GetToken() != TOKEN_KEYWORD || GetString() != Keyword)
		CreateError(_T("Expected keyword: ") + Keyword);
}

const tchar * Tokenizer::GetTokenName(TOKEN Token)
{
	switch (Token)
	{
	case TOKEN_EOF:         return _T("End of data");
	case TOKEN_EOL:         return _T("End of line");
	case TOKEN_SYMBOL:      return _T("Symbol");
	case TOKEN_INTEGER:     return _T("Integer number");
	case TOKEN_FLOAT:       return _T("Float number");
	case TOKEN_CHAR:        return _T("Character constant");
	case TOKEN_IDENTIFIER:  return _T("Identifier");
	case TOKEN_KEYWORD:     return _T("Keyword");
	case TOKEN_STRING:      return _T("String");
	default:                return _T("Unknown");
	}
}

void Tokenizer::Escape(tstring *Out, const tstring &In, uint EscapeFlags)
{
	Out->clear();
	Out->reserve(In.length());

	bool EscapeEOL   = ( (EscapeFlags & ESCAPE_EOL  ) != 0 );
	bool EscapeOther = ( (EscapeFlags & ESCAPE_OTHER) != 0 );

	tchar Ch;
	for (uint i = 0; i < In.length(); i++)
	{
		Ch = In[i]; 
		if (Ch == _T('\r'))
		{
			if (EscapeEOL)
				*Out += _T("\\r");
			else
				*Out += _T('\r');
		}
		else if (Ch == _T('\n'))
		{
			if (EscapeEOL)
				*Out += _T("\\n");
			else
				*Out += _T('\n');
		}
		else if (Ch == _T('\\') || Ch == _T('\'') || Ch == _T('\"'))
		{
			*Out += '\\';
			*Out += Ch;
		}
		else if (Ch == _T('\0')) *Out += _T("\\0");
		else if (Ch == _T('\v')) *Out += _T("\\v");
		else if (Ch == _T('\b')) *Out += _T("\\b");
		else if (Ch == _T('\f')) *Out += _T("\\f");
		else if (Ch == _T('\a')) *Out += _T("\\a");
		else if (Ch == _T('\t')) *Out += Ch;
		else if ( ((uint1)Ch < 32 || (uint1)Ch > 126) && EscapeOther )
		{
			*Out += _T("\\x");
			tstring s;
			UintToStr2(&s, (uint1)Ch, 2, 16);
			*Out += s;
		}
		else
			*Out += Ch;
	}
}

} // namespace common
