/** \file
\brief Parser skadni jzyka podobnego do C/C++
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_Tokenizer \n
Elementy moduu: \ref code_tokenizer
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_TOKENIZER_H_
#define COMMON_TOKENIZER_H_

#include "Error.hpp"

namespace common
{

class Stream;

/** \addtogroup code_tokenizer Modu Tokenizer
Dokumentacja: \ref Module_Tokenizer \n
Nagwek: Tokenizer.hpp */
//@{

/// Klasa bdu dla tokenizera
class TokenizerError : public Error
{
private:
	size_t m_Char, m_Row, m_Col;
public:
	TokenizerError(size_t Char, size_t Row, size_t Col, const tstring &Msg = _T(""), const tstring &File = _T(""), int a_Line = 0);

	size_t GetCharNum() { return m_Char; }
	size_t GetRowNum() { return m_Row; }
	size_t GetColNum() { return m_Col; }
};

/// \internal
class Tokenizer_pimpl;

/// Tokenizer
class Tokenizer
{
private:
	scoped_ptr<Tokenizer_pimpl> pimpl;

public:
	/// Koniec wiersza '\n' ma by uznawany za osobnego rodzaju token, nie za zwyczajny biay znak
	static const uint FLAG_TOKEN_EOL         = 0x01;
	/// acuch "" moe si rozciga na wiele wierszy zawierajc w treci koniec wiersza
	static const uint FLAG_MULTILINE_STRINGS = 0x02;

	/// Rodzaje tokenw
	enum TOKEN
	{
		TOKEN_EOF,        ///< Koniec caego dokumentu (dalej nie parsowa)
		TOKEN_EOL,        ///< Token koca linii (tylko kiedy FLAG_TOKEN_EOL)
		TOKEN_SYMBOL,     ///< Jednoznakowy symbol (np. kropka)
		TOKEN_INTEGER,    ///< Liczba cakowita, np. <tt>-123</tt>
		TOKEN_FLOAT,      ///< Liczba zmiennoprzecinkowa, np. <tt>10.5</tt>
		TOKEN_CHAR,       ///< Staa znakowa, np. <tt>'A'</tt>
		TOKEN_IDENTIFIER, ///< Identyfikator, np. <tt>abc_123</tt>
		TOKEN_KEYWORD,    ///< Sowo kluczowe, czyli identyfikator pasujcy do zarejestrowanych sw kluczowych
		TOKEN_STRING,     ///< Staa acuchowa, np. <tt>"ABC"</tt>
	};

	/// Tworzy z dokumentu podanego przez acuch tchar*
	Tokenizer(const tchar *Input, size_t NumChars, uint Flags);
	/// Tworzy z dokumentu podanego przez tstring
	Tokenizer(const tstring *Input, uint Flags);
	/// Tworzy z dokumentu wczytywanego z dowolnego strumienia
	Tokenizer(Stream *Input, uint Flags);

	~Tokenizer();

	/** \name Konfiguracja */
	//@{
	void RegisterKeyword(uint Id, const tstring &Keyword);
	/// Wczytuje ca list sw kluczowych w postaci tablicy acuchw tchar*.
	/** Sowa kluczowe bd miay kolejne identyfikatory 0, 1, 2 itd. */
	void RegisterKeywords(const tchar **Keywords, size_t KeywordCount);
	//@}

	/** \name Funkcje podstawowe
	Zwracaj informacje na temat ostatnio odczytanego tokena. */
	//@{

	/// Odczytuje nastpny token
	void Next();

	/// Zwraca typ tokena
	TOKEN GetToken();

	size_t GetCharNum();
	size_t GetRowNum();
	size_t GetColNum();

	/// Dziaa zawsze, ale zastosowanie gwnie dla GetToken() == common::Tokenizer::TOKEN_IDENTIFIER lub common::Tokenizer::TOKEN_STRING
	const tstring & GetString();
	/// Tylko jeli GetToken() == common::Tokenizer::TOKEN_CHAR lub common::Tokenizer::TOKEN_SYMBOL
	tchar GetChar();
	/// Tylko jeli GetToken() == common::Tokenizer::TOKEN_KEYWORD
	uint GetId();
	/// Tylko jeli GetToken() == common::Tokenizer::TOKEN_INTEGER
	bool GetUint1(uint1 *Out);
	bool GetUint2(uint2 *Out);
	bool GetUint4(uint4 *Out);
	bool GetUint8(uint8 *Out);
	bool GetInt1(int1 *Out);
	bool GetInt2(int2 *Out);
	bool GetInt4(int4 *Out);
	bool GetInt8(int8 *Out);
	uint1 MustGetUint1();
	uint2 MustGetUint2();
	uint4 MustGetUint4();
	uint8 MustGetUint8();
	int1 MustGetInt1();
	int2 MustGetInt2();
	int4 MustGetInt4();
	int8 MustGetInt8();
	/** Tylko jeli GetToken() == common::Tokenizer::TOKEN_INTEGER lub common::Tokenizer::TOKEN_FLOAT */
	bool GetFloat(float *Out);
	/** Tylko jeli GetToken() == common::Tokenizer::TOKEN_INTEGER lub common::Tokenizer::TOKEN_FLOAT */
	bool GetDouble(double *Out);
	float MustGetFloat();
	double MustGetDouble();

	//@}

	/// Rzuca wyjtek z bdem parsowania. Domylny komunikat.
	void CreateError();
	/// Rzuca wyjtek z bdem parsowania. Wasny komunikat.
	void CreateError(const tstring &Msg);

	/** \name Zapytania
	Zwracaj true, jeli ostatnio sparsowany token jest taki jak tu podany. */
	//@{
	bool QueryToken(TOKEN Token);
	bool QueryToken(TOKEN Token1, TOKEN Token2);
	bool QueryEOF();
	bool QueryEOL();
	bool QuerySymbol(tchar Symbol);
	bool QueryIdentifier(const tstring &Identifier);
	bool QueryKeyword(uint KeywordId);
	bool QueryKeyword(const tstring &Keyword);
	//@}

	/** \name Asercje
	Rzucaj wyjtek, jeli ostatnio wczytany token nie spenia danego warunku. */
	//@{
	/// Rzuca wyjtek z bdem parsowania, jeli ostatnio sparsowany token nie jest podanego rodzaju.
	void AssertToken(TOKEN Token);
	/// Rzuca wyjtek z bdem parsowania, jeli ostatnio sparsowany token nie jest jednego z dwch podanych rodzajw.
	void AssertToken(TOKEN Token1, TOKEN Token2);
	void AssertEOF();
	void AssertEOL();
	void AssertSymbol(tchar Symbol);
	void AssertIdentifier(const tstring &Identifier);
	void AssertKeyword(uint KeywordId);
	void AssertKeyword(const tstring &Keyword);
	//@}

	/// Zamienia typ tokena na jego czytelny dla uytkownika opis tekstowy, np. "Staa znakowa"
	static const tchar * GetTokenName(TOKEN Token);

	static const uint ESCAPE_EOL   = 0x01; ///< Zamienia znaki koca wiersza na <tt>\\r \\n</tt> na sekwencje ucieczki
	static const uint ESCAPE_OTHER = 0x02; ///< Zamienia znaki niedostne z klawiatury (w tym polskie litery) na <tt>\\xHH</tt>
	/// Escapuje acuch zgodnie ze skadni tokenizera
	/** \param EscapeFlags Uywa staych ESCAPE_EOL, ESCAPE_OTHER. */
	static void Escape(tstring *Out, const tstring &In, uint EscapeFlags);
};

//@}
// code_tokenizer

} // namespace common

#endif
