/** \file
\brief Modu wspmagajcy biblioteki zlib.
\author Adam Sawicki - sawickiap@poczta.onet.pl - http://regedit.gamedev.pl/ \n

Cz biblioteki CommonLib \n
Kodowanie Windows-1250, koniec wiersza CR+LF, test: Za gl ja \n
Licencja: GNU GPL. \n
Dokumentacja: \ref Module_ZlibUtils\n
Elementy moduu: \ref code_zlibutils
*/
#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#pragma once
#endif
#ifndef COMMON_ZLIB_UTILS_H_
#define COMMON_ZLIB_UTILS_H_

#include "Stream.hpp"
#include "Error.hpp"

namespace common
{

/** \addtogroup code_zlibutils Modu ZlibUtils
Dokumentacja: \ref Module_ZlibUtils \n
Nagwek: ZlibUtils.hpp */
//@{

/// Klasa wyjtku dla bdu zgoszonego przez bibliotek zlib
class ZlibError : public Error
{
public:
	ZlibError(int Code, const tstring &Msg = _T(""), const tstring &File = _T(""), int Line = 0);
};

/** \name Poziomy kompresji Zlib. */
//@{
/// Brak kompresji (co nie znaczy e dane zostan dosownie przepisane)
const int ZLIB_STORE_LEVEL   = 0;
/// Najsabsza kompresja, ale najszybsza
const int ZLIB_FASTEST_LEVEL = 1;
/// Najlepsza jako kompresji
const int ZLIB_BEST_LEVEL    = 9;
/// Domylny poziom kompresji
extern const int ZLIB_DEFAULT_LEVEL;
//@}

/// \internal
class ZlibCompressionStream_pimpl;
/// \internal
class ZlibDecompressionStream_pimpl;

/// Strumie do zapisu, ktry kompresuje dane w formacie zlib.
class ZlibCompressionStream : public OverlayStream
{
private:
	scoped_ptr<ZlibCompressionStream_pimpl> pimpl;

public:
	ZlibCompressionStream(Stream *a_Stream, int Level = ZLIB_DEFAULT_LEVEL);
	virtual ~ZlibCompressionStream();

	// ======== Implementacja Stream ========
	
	virtual void Write(const void *Data, size_t Size);
	/** Moe spowodowa, e skompresowane dane bd inaczej wygaday i obniy jako kompresji. */
	virtual void Flush();
	
	/// Oblicza maksymalny rozmiar bufora na skompresowane dane
	static size_t CompressLength(size_t DataLength);
	/// Po prostu kompresuje podane dane
	/**
	\return Zwraca dugo skompresowanych danych.
	\param[out] OutData Musi by buforem o dugoci conajmniej takiej, jak obliczona przez CompressLength.
	\param BufLength Dugo zaalokowanej pamici na OutData, podawana i sprawdzana na wszelki wypadek.
	*/
	static size_t Compress(void *OutData, size_t BufLength, const void *Data, size_t DataLength, int Level = ZLIB_DEFAULT_LEVEL);
};

/// Strumie do zapisu, ktry kompresuje w formacie gzip doczajc nagwek gzip z podanymi parametrami.
/** Wszystko jak w klasie ZlibCompressionStream. */
class GzipCompressionStream : public OverlayStream
{
private:
	scoped_ptr<ZlibCompressionStream_pimpl> pimpl;

public:
	/**
	\param FileName Jeli ma nie by, mona poda NULL.
	\param Comment Jeli ma nie by, mona poda NULL.
	*/
	GzipCompressionStream(Stream *a_Stream, const string *FileName, const string *Comment, int Level = ZLIB_DEFAULT_LEVEL);
	virtual ~GzipCompressionStream();

	// ======== Implementacja Stream ========
	
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush();
};

/// Strumie do odczytu dekompresujcy dane w formacie Zlib.
class ZlibDecompressionStream : public OverlayStream
{
private:
	scoped_ptr<ZlibDecompressionStream_pimpl> pimpl;

public:
	ZlibDecompressionStream(Stream *a_Stream);
	virtual ~ZlibDecompressionStream();

	// ======== Implementacja Stream ========
	
	virtual size_t Read(void *Out, size_t MaxLength);
	virtual bool End();

	/// Po prostu dekompresuje podane dane
	/**
	Rozmiar danych po dekompresji musi by gdzie zapamitany i odpowiednio duy bufor musi by ju zaalkowany.
	\return Zwraca dugo rozkompresowanych danych.
	\param BufLength Dugo zaalokowanej pamici na OutData, podawana i sprawdzana na wszelki wypadek.
	*/
	static size_t Decompress(void *OutData, size_t BufLength, const void *Data, size_t DataLength);
};

/// Strumie do odczytu, ktry dekompresuje dane w formacie gzip odczytujc te nagwek.
/** Wszystko podobnie jak w klasie ZlibCompressionStream. */
class GzipDecompressionStream : public OverlayStream
{
private:
	scoped_ptr<ZlibDecompressionStream_pimpl> pimpl;

public:
	GzipDecompressionStream(Stream *a_Stream);
	virtual ~GzipDecompressionStream();

	// ======== Implementacja Stream ========
	
	virtual size_t Read(void *Out, size_t MaxLength);
	virtual bool End();

	/** \name Pobieranie informacji z nagwka.
	Jeli plik posiada nagwek, ten nagwek przechowuje dan informacj i
	zosta ju w toku odczytywania odczytany, zwraca true i zwraca przez parametr
	z t informacj. */
	//@{
	bool GetHeaderFileName(string *OutFileName);
	bool GetHeaderComment(string *OutComment);
	//@}
};

/// Tryb otwarcia pliku GZip
enum GZIP_FILE_MODE
{
	GZFM_WRITE, ///< Do zapisu
	GZFM_READ,  ///< Do odczytu
};

/// \internal
class GzipFileStream_pimpl;

/// Obsuga pliku w formacie gzip (zalecane rozszerzenie ".gz")
class GzipFileStream : public Stream
{
private:
	scoped_ptr<GzipFileStream_pimpl> pimpl;

public:
	/** Level ma znaczenie tylko przy zapisie. */
	GzipFileStream(const string &FileName, GZIP_FILE_MODE Mode, int Level = ZLIB_DEFAULT_LEVEL);
	virtual ~GzipFileStream();

	// ======== Implementacja Stream ========
	
	virtual void Write(const void *Data, size_t Size);
	virtual void Flush();
	
	virtual size_t Read(void *Out, size_t MaxLength);
	virtual bool End();
};

//@}
// code_zlibutils

} // namespace common

#endif
